import sys
from argparse import ArgumentParser
from ucca.constructions import extract_candidates, add_argument
import xml.etree.ElementTree as ETree
from ucca import layer0, layer1, convert, textutil
from operator import attrgetter, itemgetter
from ucca.convert import split2sentences
from ucca.ioutil import get_passages, get_passages_with_progress_bar, external_write_mode
import re

# Script by Sriram Chaudhury


descr = {'Q':'Unknown', 'T':'Unknown', 'Terminal':'Terminal_node','P':'Process' ,'S':'State', 'A':'Participant', 'D':'Adverbial', 'C':'Center', 'E':'Elaborator', 'N':'Connector', 'R':'Relator', 'H':'Parallel_Scene', 'L':'Linker' ,'G':'Ground', 'F':'Function', 'U':'Punctuation'
}

def find_path(node,path):
        if(len(node.parents) >= 1):
            if(node.tag != 'Word') and (node.tag != 'Punctuation'):
                #path.append(node.ID+'--'+node.ftag+':'+descr[node.ftag]+'-->'+node.parents[0].ID)
                path.append(('-->('+node.ftag+':'+descr[node.ftag]+')-->'+node.parents[0].ID))
            else:
                #path.append(node.text+'--'+'Terminal'+'-->'+node.parents[0].ID)
                path.append((node.text+'('+str(node.ID)+')'+'--Terminal-->'+node.parents[0].ID))
            for j in node.parents:
                find_path(j,path)
        return path


def find_children(node, path, level):
    remote_found = 0
    for edge in node:
        if(edge.attrib.get('remote')):
            t12 = edge
            remote_found = 1

    for ch in node.children:
        if (ch.tag != 'Word') and (ch.tag != 'Punctuation'):
            if(remote_found) and (ch.ID == t12.child.ID):
                path.append((ch.ftag, ch.ID+'*', level+1,True))
            else:
                path.append((ch.ftag, ch.ID, level+1,False))
            find_children(ch, path, level+1)
        else:
            path.append((ch.text, ch.ID, level+1,False))
            path.append('End')

    return path




def main(args):

    for passage in get_passages_with_progress_bar(args.passages):
        t = split2sentences(passage)
        sen_no = 0
        for sen in t:
            #print('sentence %d\n\n%s\n%s' %(i,convert.to_text(sen), convert.to_sequence(sen)))
            print('sentence %d\n\n%s\n' %(sen_no, convert.to_text(sen)))

            root = sen.nodes['1.1']
            first = 1
            tab_len = {}
            tab_len[0] = len('1.1')
            for i in root.children:
                print('\n')
                path = []
                level = 1
                path.append((i.ftag, i.ID, level,False))
                path = find_children(i, path, level)
                end = 0
                if(first):
                    pstr = root.ID
                    first = 0
                else:
                    for k in range(0, tab_len[0]):
                        pstr = pstr + ' '
                for j in path:
                    if(j == 'End'):
                        print(pstr)
                        pstr = ''
                        end = 1
                        continue
                    rel = j[0]
                    nd = j[1]
                    tab = int(j[2])
                    remote = j[3]
                    if(end):
                        q_mark = 0
                        for k in range(0,tab_len[tab-1]):
                            if(k == tab_len[q_mark]):
                                pstr = pstr + '.'
                                q_mark += 1
                            else:
                                pstr = pstr+' '
                            end = 0
                    if(rel in descr):
                        rel_desc = rel+':' +descr[rel]
                    else:
                        rel_desc = rel
                    if(remote):
                        pstr = pstr + '|-->Remote(' + rel_desc + ')-->' + nd
                    else:
                        pstr = pstr+'|-->('+rel_desc+')-->'+nd
                    tab_len[tab] = len(pstr)

            print('-----------------------------------\n')
            sen_no += 1



if __name__ == "__main__":
    argparser = ArgumentParser(description="Xml to conll and find the path of the word from UCCA xml file.")
    argparser.add_argument("passages", nargs="+", help="the corpus, given as xml/pickle file names")
    add_argument(argparser, False)
    main(argparser.parse_args())



#!/usr/bin/env python3
import sys

import argparse

from uccaapp.create_annotation_tasks import ServerAccessor, AnnotationTaskCreator

desc = """Upload a list of tokenization tasks to a project"""


class TokenizationTaskCreator(AnnotationTaskCreator):
    def __init__(self, project_id, **kwargs):
        super().__init__(**kwargs)
        self.set_project(project_id)

    def build_task(self, user_id, passage_id, **kwargs):
        del kwargs
        user = self.get_user(user_id)
        passage = self.get_passage(passage_id)
        return dict(type="TOKENIZATION", project=self.project, user=user, passage=passage,
                    manager_comment="passage #%s" % passage["id"], user_comment="", parent=None, is_demo=False,
                    is_active=True)

    @staticmethod
    def add_arguments(argparser):
        argparser.add_argument("filename", help="a file where each line is a <User ID> <Passage ID>")
        argparser.add_argument("-l", "--log", help="filename to write log of uploaded passages to")
        ServerAccessor.add_project_id_argument(argparser)
        #ServerAccessor.add_user_id_argument(argparser)
        ServerAccessor.add_arguments(argparser)


def main(**kwargs):
    TokenizationTaskCreator(**kwargs).create_tasks(**kwargs)


if __name__ == "__main__":
    argument_parser = argparse.ArgumentParser(description=desc)
    TokenizationTaskCreator.add_arguments(argument_parser)
    main(**vars(argument_parser.parse_args()))
    sys.exit(0)



#!/usr/bin/env python3

import argparse
import os
import sys
from itertools import count
from logging import warning

from ucca.convert import split2sentences, split_passage
from ucca.ioutil import passage2file, get_passages_with_progress_bar, external_write_mode
from ucca.normalization import normalize
from ucca.textutil import extract_terminals

desc = """Parses XML files in UCCA standard format, and writes a passage per sentence."""

NUM_NODES_WARNING = 500  # Warn if a sentence has more than this many nodes


class Splitter:
    def __init__(self, sentences, enum=False, suffix_format="%03d", suffix_start=0):
        self.sentences = sentences
        self.sentence_to_index = {}
        for i, sentence in enumerate(sentences):
            self.sentence_to_index.setdefault(sentence, []).append(i)
        self.enumerate = enum
        self.suffix_format = suffix_format
        self.suffix_start = suffix_start
        self.index = 0
        self.matched_indices = set()

    @classmethod
    def read_file(cls, filename, **kwargs):
        if filename is None:
            return None
        with open(filename, encoding="utf-8") as f:
            sentences = [line.strip() for line in f]
        return cls(sentences, **kwargs)

    def split(self, passage):
        ends = []
        ids = []
        token_lists = []
        for terminal in extract_terminals(passage):
            token_lists.append([])
            for terminals in token_lists if self.index is None else [token_lists[0]]:
                terminals.append(terminal)
                sentence = " ".join(t.text for t in terminals)
                if self.index is not None and self.index < len(self.sentences) and self.sentences[
                        self.index].startswith(sentence):  # Try matching next sentence rather than shortest
                    index = self.index if self.sentences[self.index] == sentence else None
                else:
                    indices = self.sentence_to_index.get(sentence)
                    index = self.index = indices.pop(0) if indices else None
                if index is not None:
                    self.matched_indices.add(index)
                    last_end = terminals[0].position - 1
                    if len(terminals) > 1 and last_end and last_end not in ends:
                        ends.append(last_end)
                    ends.append(terminal.position)
                    ids.append(str(index))
                    token_lists = []
                    self.index += 1
                    break
        return split_passage(passage, ends, ids=ids if self.enumerate else None,
                             suffix_format=self.suffix_format, suffix_start=self.suffix_start)


def main(args):
    splitter = Splitter.read_file(args.sentences, enum=args.enumerate,
                                  suffix_format=args.suffix_format, suffix_start=args.suffix_start)
    os.makedirs(args.outdir, exist_ok=True)
    i = 0
    for passage in get_passages_with_progress_bar(args.filenames, "Splitting"):
        for sentence in splitter.split(passage) if splitter else split2sentences(
                passage, remarks=args.remarks, lang=args.lang, ids=map(str, count(i)) if args.enumerate else None):
            i += 1
            outfile = os.path.join(args.outdir, args.prefix + sentence.ID + (".pickle" if args.binary else ".xml"))
            if len(sentence.nodes) > NUM_NODES_WARNING:
                warning(f"Sentence {i} in passage {passage.ID} has {len(sentence.nodes)} > {NUM_NODES_WARNING} nodes")
            if args.verbose:
                with external_write_mode():
                    print(sentence, file=sys.stderr)
                    print("Writing passage file for sentence '%s'..." % outfile, file=sys.stderr)
            if args.normalize:
                normalize(sentence)
            passage2file(sentence, outfile, binary=args.binary)
    if splitter and len(splitter.matched_indices) < len(splitter.sentences):
        print("", "Unmatched sentences:", *[s for i, s in enumerate(splitter.sentences)
                                            if i not in splitter.matched_indices], sep="\n")


if __name__ == "__main__":
    argparser = argparse.ArgumentParser(description=desc)
    argparser.add_argument("filenames", nargs="+", help="passage file names to convert")
    argparser.add_argument("-o", "--outdir", default=".", help="output directory")
    argparser.add_argument("-p", "--prefix", default="", help="output filename prefix")
    argparser.add_argument("-f", "--suffix-format", default="%03d", help="sentence number suffix format")
    argparser.add_argument("-i", "--suffix-start", type=int, default=0, help="start index for number suffix")
    argparser.add_argument("-r", "--remarks", action="store_true", help="annotate original IDs")
    argparser.add_argument("-l", "--lang", default="en", help="language two-letter code for sentence model")
    argparser.add_argument("-b", "--binary", action="store_true", help="write in pickle binary format (.pickle)")
    argparser.add_argument("-s", "--sentences", help="optional input file with sentence at each line to split by")
    argparser.add_argument("-e", "--enumerate", action="store_true", help="set each output sentence ID by global order")
    argparser.add_argument("-N", "--no-normalize", dest="normalize", action="store_false",
                           help="do not normalize passages after splitting")
    argparser.add_argument("-v", "--verbose", action="store_true", help="print information about every split sentence")
    main(argparser.parse_args())



"""This module encapsulate the basic elements of the UCCA annotation.

A UCCA annotation is practically a directed acyclic graph (DAG), which
represents a :class:`Passage` of text and its annotation. The annotation itself
is divided into :class:`Layer` objects, where in each layer :class:`Node` objects
are connected between themselves and to Nodes in other layers using
:class:`Edge` objects.

"""

import functools

# Max number of digits allowed for a unique ID
UNIQUE_ID_MAX_DIGITS = 5

# Attribute to ignore when comparing entities
IRRELEVANT_ATTRIBUTES = {"uncertain"}


# Used as the default ordering key function for ordered objects, namely
# :class:`Layer` and :class:`Node` .
def id_orderkey(node):
    """Key function which sorts by layer (string), then by unique ID (int).

    Args:
        node: :class:`Node` which we will to sort according to its ID

    Returns:
        a string with the layer and unique ID in such a way that sort will
        first order lexicography the layer ID then numerically the unique ID.

    """
    layer, unique = node.ID.split(Node.ID_SEPARATOR)
    return "{} {:>{}}".format(layer, unique, UNIQUE_ID_MAX_DIGITS)


def edge_id_orderkey(edge):
    """Key function which sorts Edges by its IDs (using :func:`id_orderkey`).

    Args:
        edge: :class:`Edge` which we wish to sort according to the ID of its
        parent and children after using :func:`id_orderkey`.

    Returns:
        a string with the layer and unique ID in such a way that sort will
        first order lexicography the layer ID then numerically the unique ID.

    """
    return Edge.ID_FORMAT.format(id_orderkey(edge.parent),
                                 id_orderkey(edge.child))


class UCCAError(Exception):
    """Base class for all UCCA package exceptions."""
    pass


class FrozenPassageError(UCCAError):
    """Exception raised when trying to modify a frozen :class:`Passage`."""
    pass


class DuplicateIdError(UCCAError):
    """Exception raised when trying to add an element with an existing ID.

    For each element, a unique ID must be assigned. If the ID of the new element
    is already present in the :class:`Passage` in some way, this exception is
    raised.

    """
    pass


class MissingNodeError(UCCAError):
    """Exception raised when trying to access a non-existent :class:`Node`."""
    pass


class UnimplementedMethodError(UCCAError):
    """Exception raised when trying to call a not-yet-implemented method."""
    pass


class ModifyPassage:
    """Decorator for changing a :class:`Passage` or any member of it.

    This decorator is mandatory for anything which causes the elements in
    a :class:`Passage` to change by adding or removing an element, or changing
    an attribute.

    It validates that the Passage is not frozen before allowing the change.

    The decorator can't be used for __init__ calls, as at the stage of the
    check there are no instance attributes to check. So in such cases,
    a function that binds the object created with the Passage should be
    decorated instead (and should be called after the instance attributes
    are set).

    Attributes:
        fn: the function object to decorate

    """

    def __init__(self, fn):
        self.fn = fn

    def __get__(self, obj, cls):
        """Used to bind the function to the instance (add 'self')."""
        return functools.partial(self.__call__, obj)

    def __call__(self, *args, **kwargs):
        """Decorating functions which modify :class:`Passage` elements.

        :param args: list of all arguments, assuming the first is the object
                which modifies :class:`Passage`, and it has an attribute root
                which points to the Passage it is part of.
        :param kwargs: list of all keyword arguments
        :return: The decorated function result.
        :raise FrozenPassageError: if the :class:`Passage` is frozen and can't be
                modified.
        """

        @functools.wraps(self.fn)
        def decorated(*args, **kwargs):
            if args[0].root.frozen:
                raise FrozenPassageError(args[0].root.ID)
            return self.fn(*args, **kwargs)

        return decorated(*args, **kwargs)


class _AttributeDict:
    """Dictionary which stores attributes for any UCCA element.

    This dictionary is used to store attributes which are part of any
    element in the UCCA annotation scheme. It's advantage over regular
    dictionary is adhering to :class:`Passage` frozen status and modification
    decorators.

    Attributes:
        root: the Passage this object is linked with

    """

    def __init__(self, root, mapping=None):
        self._root = root
        self._dict = mapping.copy() if mapping is not None else dict()

    def __getitem__(self, key):
        return self._dict[key]

    def get(self, key, default=None):
        return self._dict.get(key, default)

    def equals(self, other):
        """True iff the two objects are equal (only dicts, w.o.r.t Passage).

        :param other: AttributeDict to compare to
        :return: True iff the dictionaries are equal.
        """

        def omit_irrelevant(d):
            return {k: v for k, v in d.items() if k not in IRRELEVANT_ATTRIBUTES}

        return omit_irrelevant(self._dict) == omit_irrelevant(other._dict)

    @property
    def root(self):
        return self._root

    def copy(self):
        return self._dict.copy()

    @ModifyPassage
    def __setitem__(self, key, value):
        self._dict[key] = value

    @ModifyPassage
    def update(self, values):
        self._dict.update(values)

    @ModifyPassage
    def __delitem__(self, key):
        del self._dict[key]

    def __len__(self):
        return len(self._dict)

    def items(self):
        return self._dict.items()


class Category:
    """when considering refinement layers, each edge can have multiple tags sorted in a certain hierarchy.
    for this reason, a category must include not only the tag information but also the layer and hierarchy
    information.
    """

    def __init__(self, tag, slot=None, layer=None, parent=None):
        self._tag = tag
        self._slot = slot if slot else ""
        self._layer = layer if layer else ""
        self._parent = parent if parent else ""
        self.extra = {}

    @property
    def tag(self):
        return self._tag

    @tag.setter
    def tag(self, new_tag):
        self._tag = new_tag

    @property
    def slot(self):
        return self._slot

    @property
    def layer(self):
        return self._layer

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, new_parent):
        self._parent = new_parent

    def to_xml(self):
        pass

    def __iter__(self):
        return iter((self.tag, self.slot, self.layer, self.parent))


class Edge:
    """Labeled edge between two :class:`Node` objects in UCCA annotation graph.

    An edge between Nodes in a :class:`Passage` is a simple object; it is a
    directed edge whose ID is derived by the parent and child of the edge,
    it is mostly immutable except for its attributes, and it is labeled with
    the connection type between the Nodes. An edge can have multiple annotations, representing
    connection types of one or more layers.

    Attributes:
        ID: ID of the Edge, constructed from the IDs of the two Nodes
        root: the Passage this object is linked with
        attrib: attribute dictionary of the Edge
        extra: temporary storage space for undocumented attributes and data
        tag: the string label of the Edge
        parent: the originating Node of the Edge
        child: the target Node of the Edge
        categories: a list of categories for this edge
        ID_FORMAT: format string which creates the ID of the Edge from
            the IDs of the parent (first argument to the formatting string)
            and the child (second argument).

    """

    ID_FORMAT = "{}->{}"

    def __init__(self, root, parent, child, tag=None, attrib=None):
        """Creates a new :class:`Edge` object.

        :param see :class:`Edge` documentation.

        :raise FrozenPassageError: if the :class:`Passage` object we are part of
                is frozen and can't be modified.

        """
        if root.frozen:
            raise FrozenPassageError(root.ID)
        self._root = root
        self._parent = parent
        self._child = child
        self._attrib = _AttributeDict(root, attrib)
        self._categories = [Category(tag)] if tag else []
        self.extra = {}

    @property
    def tag(self):
        return self.categories[0].tag

    @tag.setter
    @ModifyPassage
    def tag(self, new_tag):
        old_tag = self.tag
        self.categories[0].tag = new_tag
        self._root._change_edge_tag(self, old_tag)

    @property
    def tags(self):
        return [category.tag for category in self.categories]

    @property
    def root(self):
        return self._root

    @property
    def parent(self):
        return self._parent

    @property
    def categories(self):
        try:
            return self._categories
        except AttributeError as e:
            raise IOError("Load passage %s from pickle saved with UCCA<1.1. Please load from XML instead." %
                          self.root.ID) from e

    @categories.setter
    def categories(self, new_categories):
        self._categories = new_categories

    @property
    def child(self):
        return self._child

    @property
    def attrib(self):
        return self._attrib

    @property
    def ID(self):
        return Edge.ID_FORMAT.format(self._parent.ID, self._child.ID)

    def equals(self, other, *, recursive=True, ordered=False,
               ignore_node=None, ignore_edge=None):
        """Returns whether self and other are Edge-equals.

        Edge-equality is determined by having the same tag and attributes.
        Recursive Edge-equality means that the Edges are equal, and their
        children are recursively Node-equal.

        :param other: an Edge object to compare to
        :param recursive: whether to compare recursively, defaults to True
        :param ordered: if recursive, whether the children are Node-equivalent
            w.r.t order (see Node.equals())
        :param ignore_node: function that returns whether to ignore a given node
        :param ignore_edge: function that returns whether to ignore a given edge


        :return: True iff the Edges are equal.

        """
        return self.tag == other.tag and \
               self._attrib.equals(other._attrib) and \
               (not recursive or
                self.child.equals(other.child,
                                  ordered=ordered,
                                  ignore_node=ignore_node, ignore_edge=ignore_edge))

    @ModifyPassage
    def add(self, tag, slot="", layer="", parent=""):
        """ adds a new category to the edge"""
        c = Category(tag, slot, layer, parent)
        self.categories.append(c)
        if c.tag not in self.root.categories:
            self.root._update_categories(c)
        if c.parent and c.parent not in self.root.refined_categories:
            self.root._update_refined_categories(c.parent)
        return c

    def __repr__(self):
        return self.ID

    def __getitem__(self, index):
        return self.categories[index]


class Node:
    """Labeled Node in UCCA annotation graph.

    A Node in :class:`Passage` UCCA annotation is an vertex in the annotation
    graph, which may be an internal vertex or a leaf, and is labeled with a
    tag that specifies both the :class:`Layer` it belongs to and it's ID in this
    Layer. It can have multiple children Nodes through :class:`Edge` objects,
    and these children are ordered according to an internal order function.

    Attributes:
        ID: ID of the Node, constructed from the ID of the Layer it belongs to,
            a separator, and a unique alphanumeric ID in the layer.
        root: the Passage this object is linked with
        attrib: attribute dictionary of the Node
        extra: temporary storage space for undocumented attributes and data
        tag: the string label of the Node
        layer: the Layer this Node belongs to
        incoming: a copy of the incoming Edges to this object
        outgoing: a copy of the outgoing Edges from this object
        parents: the Nodes which have incoming Edges to this object
        children: the Nodes which have outgoing Edges from this object
        orderkey: the key function for ordering the outgoing Edges
        ID_SEPARATOR: separator function between the Layer ID and the unique
            Node ID in the complete ID of the Node. Mustn't be alphanumeric.

    """

    ID_SEPARATOR = '.'

    def __init__(self, ID, root, tag, attrib=None, *,
                 orderkey=edge_id_orderkey):
        """Creates a new :class:`Node` object.

        :param see :class:`Node` documentation.

        :raise FrozenPassageError: if the :class:`Passage` object we are part of
                is frozen and can't be modified.

        """
        if root.frozen:
            raise FrozenPassageError(root.ID)
        self._tag = tag
        self._root = root
        self._ID = ID
        self._attrib = _AttributeDict(root, attrib)
        self.extra = {}
        self._outgoing = []
        self._incoming = []
        self._orderkey = orderkey

        # After properly initializing self, add it to the Passage/Layer
        root._add_node(self)
        try:
            root.layer(self.layer.ID)._add_node(self)
        except KeyError as e:
            raise ValueError("Invalid layer '%s' in node ID '%s'" % (self.layer.ID, self._ID)) from e

    @property
    def tag(self):
        return self._tag

    @tag.setter
    @ModifyPassage
    def tag(self, new_tag):
        old_tag = self._tag
        self._tag = new_tag
        self._root._change_node_tag(self, old_tag)

    @property
    def root(self):
        return self._root

    @property
    def ID(self):
        return self._ID

    @property
    def attrib(self):
        return self._attrib

    @property
    def layer(self):
        return self._root.layer(self._ID.split(Node.ID_SEPARATOR)[0])

    @property
    def incoming(self):
        return tuple(self._incoming)

    @property
    def outgoing(self):
        return tuple(self._outgoing)

    @property
    def parents(self):
        return [edge.parent for edge in self._incoming]

    @property
    def children(self):
        return [edge.child for edge in self._outgoing]

    def __bool__(self):
        return True

    def __len__(self):
        return len(self._outgoing)

    def __getitem__(self, index):
        return self._outgoing[index]

    def __repr__(self):
        return Node.__name__ + "(" + self.ID + ")"

    @ModifyPassage
    def add_multiple(self, edge_categories, node, *, edge_attrib=None):
        """Adds another :class:`Node` object as a child of self.

        :param edge_categories: a list of 4-tuples representing the categories on this edge of the :class:`Edge`
                connecting between the Nodes
        :param node: the Node object which we want to have an Edge to
        :param edge_attrib: Keyword only, dictionary of attributes to be passed
                to the Edge initializer.

        :return: the newly created Edge object

        :raise FrozenPassageError: if the :class:`Passage` object we are part of
                is frozen and can't be modified.

        """
        edge = Edge(root=self._root, parent=self,
                    child=node, attrib=edge_attrib)
        for category in edge_categories:
            edge.add(*category)
        self._outgoing.append(edge)
        self._outgoing.sort(key=self._orderkey)
        node._incoming.append(edge)
        node._incoming.sort(key=node._orderkey)
        self.root._add_edge(edge)
        return edge

    @ModifyPassage
    def add(self, tag, node, *, edge_attrib=None):
        """Adds another :class:`Node` object as a child of self.

        :param edge_categories: a list of 4-tuples representing the categories on this edge of the :class:`Edge`
                connecting between the Nodes
            node: the Node object which we want to have an Edge to
            edge_attrib: Keyword only, dictionary of attributes to be passed
                to the Edge initializer.

        :return: the newly created Edge object

        :raise FrozenPassageError: if the :class:`Passage` object we are part of
                is frozen and can't be modified.

        """
        return self.add_multiple([(tag,)], node, edge_attrib=edge_attrib)

    @ModifyPassage
    def remove(self, edge_or_node):
        """Removes the :class:`Edge` between self and a child :class:`Node`.

        This methods removes the Edge given, or the Edge connecting self and
        the Node given, from the annotation of :class:`Passage`. It does not
        remove the target or originating Node from the graph but just unlinks
        them.

        :param edge_or_node: either an Edge or Node object to remove/unlink

        :raise MissingNodeError: if the Node or Edge is not connected with self.

        """
        if edge_or_node not in self._outgoing:  # a Node, or an error
            try:
                edge = [edge for edge in self._outgoing
                        if edge.child == edge_or_node][0]
            except IndexError as e:
                raise MissingNodeError(edge_or_node) from e
        else:  # an Edge object
            edge = edge_or_node

        try:
            self._outgoing.remove(edge)
            edge.child._incoming.remove(edge)
            self.root._remove_edge(edge)
        except ValueError as e:
            raise MissingNodeError(edge_or_node) from e

    @property
    def orderkey(self):
        return self._orderkey

    @orderkey.setter
    def orderkey(self, value):
        self._orderkey = value
        self._outgoing.sort(key=value)

    @ModifyPassage
    def destroy(self):
        """Removes the :class:`Node` from the :class:`Passage` annotation graph.

        This method unlinks self from all other :class:`Node` objects and removes
        self from the :class:`Layer` and Passage objects.

        """
        # using outgoing and incoming so I won't change the list I'm working on
        for edge in self.outgoing:
            self.remove(edge)
        for edge in self.incoming:
            edge.parent.remove(edge)
        self.layer._remove_node(self)
        self._root._remove_node(self)

    def equals(self, other, *, recursive=True, ordered=False,
               ignore_node=None, ignore_edge=None):
        """Returns whether the self Node-equals other.

        Node-equality is basically determined by self and other having the same
        tag and attributes. Recursive equality is achieved when all outgoing
        Edges are Edge-equal, and their children are recursively Node-equal
        as well. Ordered equality means that the outgoing Edges should be
        equivalent to each other w.r.t order (the first to the first etc.),
        while unordered equality means that each Edges are equivalent after
        being ordered with some determined order.

        :param other: the Node object to compare to
        :param recursive: whether comparison is recursive, defaults to True.
        :param ordered: whether comparison should include strict ordering
        :param ignore_node: function that returns whether to ignore a given node
        :param ignore_edge: function that returns whether to ignore a given edge

        :return: True iff the Nodes are equal in the terms given.

        """
        if self.tag != other.tag or not self._attrib.equals(other._attrib):
            return False
        if not recursive:
            return True
        edges, other_edges = [[edge for edge in node
                               if (ignore_node is None or
                                   not ignore_node(edge.child)) and (
                                       ignore_edge is None or
                                       not ignore_edge(edge))]
                              for node in (self, other)]
        if len(edges) != len(other_edges):
            return False  # not necessary, but gives better performance
        if ordered:
            return all(e1.equals(e2, ordered=True,
                                 ignore_node=ignore_node, ignore_edge=ignore_edge)
                       for e1, e2 in zip(edges, other_edges))
        # For unordered equality, I try to find & remove an equivalent
        # Edge + Node couple from other's Edges until exhausted.
        # Because both Edge-equality and Node-equality are equivalence
        # classes, I can just take the first I found and remove it w/o
        # trying to iterate through possible orders.
        try:
            for e1 in edges:
                other_edges.remove(next(e2 for e2 in other_edges
                                        if e1.equals(e2, ignore_node=ignore_node,
                                                     ignore_edge=ignore_edge)))
        except StopIteration:
            return False
        return not other_edges

    def missing_edges(self, other, ignore_node=None):
        """Returns edges present in this node but missing in the other.

        :param other: the Node object to compare to
        :param ignore_node: function that returns whether to ignore a given node

        :return: List of edges present in this node but missing in the other.

        """
        edges, other_edges = [[edge for edge in node
                               if ignore_node is None or
                               not ignore_node(edge.child)]
                              for node in (self, other)]
        return sorted([e1 for e1 in edges if not any(e1.equals(e2) for e2 in other_edges)],
                      key=edge_id_orderkey)

    def iter(self, obj="nodes", method="dfs", duplicates=False, key=None):
        """Iterates the :class:`Node` objects in the subtree of self.

        :param obj: yield Node objects (use value "nodes", default) or Edge
                objects (use values "edges")
            method: do breadth-first iteration (use value "bfs") or depth-first
                iteration (value "dfs", default).
            duplicates: If True, may return the same object twice if it is
                encountered twice, because of the DAG structure which isn't
                necessarily a tree. If it is False, all objects will be yielded
                only the first time they are encountered. Defaults to False.
            key: boolean function that filters the iterable items. key function
                takes one argument (the item) and returns True if it should be
                returned to the user. If an item isn't returned, its subtree
                is still iterated.  Defaults to None (returns all items).

        Yields:
            a :class:`Node` or :class:`Edge` object according to the iteration
            parameters.

        """
        if method not in ("dfs", "bfs"):
            raise ValueError("method can be either 'dfs' or 'bfs'")
        if obj not in ("nodes", "edges"):
            raise ValueError("obj can be either 'nodes' or 'edges'")
        processed = set()
        if obj == 'nodes':
            waiting = [self]
        else:
            waiting = self._outgoing[:]
        while len(waiting):
            curr = waiting.pop(0)
            if key is None or key(curr):
                yield curr
            processed.add(curr)
            to_add = curr.children if obj == 'nodes' else list(curr.child)
            to_add = [x for x in to_add if duplicates or x not in processed]
            if method == "bfs":
                waiting.extend(to_add)
            else:
                waiting = to_add + waiting

    def get_terminals(self, *args, **kwargs):
        """Returns a list of all terminals under the span of this Node."""
        return [t for e in self._outgoing for t in e.child.get_terminals(*args, **kwargs)]


class Layer:
    """Group of similar :class:`Node` objects in UCCA annotation graph.

    A Layer in UCCA annotation graph is a subgraph of the whole :class:`Passage`
    annotation graph which consists of similar Nodes and :class:`Edge` objects
    between them. The Nodes and the Layer itself has some formal definition for
    being grouped together.

    Attributes:
        ID: ID of the Layer, must be alphanumeric.
        root: the Passage this object is linked with
        attrib: attribute dictionary of the Layer
        extra: temporary storage space for undocumented attributes and data
        orderkey: the key function for ordering the Nodes in the layer.
            Note that it must rely only on the Nodes and/or Edges in the Layer.
            If it, for example, rely on Edges added between Nodes in the Layer
            and Nodes outside the Layer (hence, the Edges are not in the Layer)
            the order will not be updated (because the Layer object won't know
            that something has changed).
        all: a list of all the Nodes which are part of this Layer
        heads: a list of all Nodes which have no incoming Edges in the subgraph
            of the Layer (can have Edges from Nodes in other Layers).

    """

    def __init__(self, ID, root, attrib=None, *, orderkey=id_orderkey):
        """Creates a new :class:`Layer` object.

        :param see :class:`Layer` documentation.

        :raise FrozenPassageError: if the :class:`Passage` object we are part of
                is frozen and can't be modified.

        """
        if root.frozen:
            raise FrozenPassageError(root.ID)
        self._ID = ID
        self._root = root
        self._attrib = _AttributeDict(root, attrib)
        self.extra = {}
        self._all = []
        self._heads = []
        self._orderkey = orderkey
        root._add_layer(self)

    @property
    def ID(self):
        return self._ID

    @property
    def root(self):
        return self._root

    @property
    def attrib(self):
        return self._attrib

    @property
    def all(self):
        return self._all[:]

    @property
    def heads(self):
        return self._heads[:]

    @property
    def orderkey(self):
        return self._orderkey

    @orderkey.setter
    def orderkey(self, value):
        self._orderkey = value
        self._all.sort(key=value)
        self._heads.sort(key=value)

    def equals(self, other, *, ordered=False, ignore_node=None, ignore_edge=None):
        """Returns whether two Layer objects are equal.

        Layers are considered Layer-equal if their attribute dictionaries are
        equal and all their heads are recursively Node-equal.
        Ordered Layer-equality implies that the heads should be
        ordered the same for the Layers to be considered equal, and the
        Node-equality is ordered too.

        :param other: the Layer object to compare to
        :param ordered: whether strict-order equality is used, defaults to False
        :param ignore_node: function that returns whether to ignore a given node
        :param ignore_edge: function that returns whether to ignore a given edge

        :return: True iff self and other are Layer-equal.

        """
        if not self._attrib.equals(other._attrib):
            return False
        heads, other_heads = [[head for head in layer.heads
                               if ignore_node is None or
                               not ignore_node(head)]
                              for layer in (self, other)]
        if len(heads) != len(other_heads):
            return False  # can be removed, here for performance gain
        if ordered:
            return all(x1.equals(x2, ordered=True,
                                 ignore_node=ignore_node, ignore_edge=ignore_edge)
                       for x1, x2 in zip(heads, other_heads))
        # I can just find the first equal head in unordered search, as
        # Node-equality is an equivalence class (see their for details).
        try:
            for h1 in heads:
                other_heads.remove(next(h2 for h2 in other_heads
                                        if h1.equals(h2, ignore_node=ignore_node,
                                                     ignore_edge=ignore_edge)))
        except StopIteration:
            return False
        return not other_heads

    def _add_edge(self, edge):
        """Alters self.heads if an :class:`Edge` has been added to the subgraph.

        Should be called when both :class:`Node` objects of the edge are part
        of this Layer (and hence part of the subgraph of it).

        :param edge: the Edge added to the Layer subgraph

        """
        if edge.child in self._heads:
            self._heads.remove(edge.child)
        # Order may depend on edges, so re-order
        self._all.sort(key=self._orderkey)
        self._heads.sort(key=self._orderkey)

    def _remove_edge(self, edge):
        """Alters self.heads if an :class:`Edge` has been removed.

        Should be called when the child :class:`Node` object of the edge is part
        of this Layer (and hence part of the subgraph of it).

        :param edge: the Edge removed from the Layer subgraph

        """
        if edge.child.layer == self and all(p.layer != self for p in edge.child.parents):
            self._heads.append(edge.child)
            self._heads.sort(key=self._orderkey)
        # Order may depend on edges, so re-order
        self._all.sort(key=self._orderkey)
        self._heads.sort(key=self._orderkey)

    def _add_node(self, node):
        """Adds a :class:`node` to the :class:`Layer`.

        Assumes node has no incoming or outgoing :class:`Edge` objects.

        """
        self._all.append(node)
        self._all.sort(key=self._orderkey)
        self._heads.append(node)
        self._heads.sort(key=self._orderkey)

    def _remove_node(self, node):
        """Removes a :class:`node` from the :class:`Layer`.

        Assumes node has no incoming or outgoing :class:`Edge` objects.

        """
        self._all.remove(node)
        self._heads.remove(node)

    def _change_edge_tag(self, edge, old_tag):
        """Updates the :class:`Layer` objects with the change.

        :param edge: the updated :class:`Edge` object
            old_tag: the Edge's tag before the change

        """
        pass  # meant to be overriden by subclasses

    def _change_node_tag(self, node, old_tag):
        """Updates the :class:`Layer` objects with the change.

        :param node: the updated :class:`Node` object
            old_tag: the Node's tag before the change

        """
        pass  # meant to be overriden by subclasses


class Passage:
    """An annotated text with UCCA annotation graph.

    A Passage is an object representing a text annotated with UCCA annotation.
    UCCA annotation is a directed acyclic graph of :class:`Node` and :class:`Edge`
    objects grouped into :class:`Layer` objects.

    Attributes:
        ID: ID of the Passage
        root: simply self, for API similarity with other UCCA objects
        attrib: attribute dictionary of the Passage
        extra: temporary storage space for undocumented attributes and data
        layers: all Layers of the Passage, no order guaranteed
        nodes: dictionary of ID-node pairs for all the nodes in the Passage
        frozen: indicates whether the Passage can be modified or not, boolean.

    """

    def __init__(self, ID, attrib=None):
        """Creates a new :class:`Passage` object.

        :param see :class:`Passage` documentation.

        """
        self._ID = ID
        self._attrib = _AttributeDict(self, attrib)
        self.extra = {}
        self._layers = {}
        self._nodes = {}
        self._categories = {}
        self._refined_categories = []
        self.frozen = False

    @property
    def ID(self):
        return self._ID

    @property
    def root(self):
        return self

    @property
    def attrib(self):
        return self._attrib

    @property
    def layers(self):
        return self._layers.values()

    @property
    def nodes(self):
        return self._nodes.copy()

    @property
    def categories(self):
        return self._categories.copy()

    @property
    def refined_categories(self):
        return self._refined_categories

    def layer(self, ID):
        """Returns the :class:`Layer` object whose ID is given.

        :param ID: ID of the Layer requested.

        :raise KeyError: if no Layer with this ID is present

        """
        try:
            return self._layers[ID]
        except KeyError as e:
            raise KeyError("Layer '%s' not found in passage '%s'" % (ID, self.ID)) from e

    def equals(self, other, *, ordered=False, ignore_node=None, ignore_edge=None):
        """Returns whether two passages are equivalent.

        Passage-equivalence is determined by having the same attributes and
        all layers (according to ID) are Layer-equivalent.

        :param other: the Passage object to compare to
        :param ordered: is Layer-equivalency should be ordered (see there)
        :param ignore_node: function that returns whether to ignore a given node
        :param ignore_edge: function that returns whether to ignore a given edge

        :return: True iff self is Passage-equivalent to other.

        """
        if not self._attrib.equals(other._attrib):
            return False
        # noinspection PyTypeChecker
        if len(self.layers) != len(other.layers):
            return False  # can be removed, here for performance gain
        try:
            for lid, l1 in self._layers.items():
                l2 = other.layer(lid)
                if not l1.equals(l2, ordered=ordered,
                                 ignore_node=ignore_node, ignore_edge=ignore_edge):
                    return False
        except KeyError:  # no layer with same ID found
            return False
        return True

    def missing_nodes(self, other, ignore_node=None, ignore_edge=None):
        """Returns nodes present in this passage but missing in the other.

        :param other: the Passage object to compare to
        :param ignore_node: function that returns whether to ignore a given node
        :param ignore_edge: function that returns whether to ignore a given edge

        :return: List of nodes present in this passage but missing in the other.

        """
        nodes, other_nodes = [[node for node in passage.nodes.values()
                               if ignore_node is None or
                               not ignore_node(node)]
                              for passage in (self, other)]
        return sorted([n1 for n1 in nodes
                       if not any(n1.equals(n2, ignore_node=ignore_node,
                                            ignore_edge=ignore_edge)
                                  for n2 in other_nodes)],
                      key=id_orderkey)

    def copy(self, layers=None):
        """Copies the Passage and specified layers to a new object.

        The main "building block" of copying is the Layer, so copying is
        truly copying the Passage attributes (attrib, extra, ID, frozen)
        and creating the equivalent layers (each layer for itself).

        :param layers: sequence of layer IDs to copy to the new object.
            If None, all layers will be copied.

        :return: A new Passage object.

        :raise KeyError: if a given layer ID doesn't exist.
            UnimplementedMethodError if copying for a layer is unimplemented.

        """
        other = Passage(ID=self.ID, attrib=self.attrib.copy())
        other.extra = self.extra.copy()
        if layers is None:
            layers = sorted(self._layers)
        for lid in layers:
            try:
                self.layer(lid).copy(other)
            except AttributeError as e:
                raise UnimplementedMethodError() from e
        other.frozen = self.frozen
        return other

    def by_id(self, ID):
        """Returns a Node whose ID is given.

        :param ID: ID string
        :return: The node.Node object whose ID matches
        :raise KeyError: if no Node with this ID is found
        """
        try:
            return self._nodes[ID]
        except KeyError as e:
            layer_id, _ = ID.split(Node.ID_SEPARATOR)
            layer = self.layer(layer_id)
            raise KeyError("Node '%s' not found in passage '%s', existing IDs in layer '%s': %s" % (
                ID, self.ID, layer_id, ", ".join(n.ID for n in layer.all))) from e

    @ModifyPassage
    def _add_layer(self, layer):
        """Adds a :class:`Layer` object to the :class:`Passage`.

        :param layer: the Layer object to add

        :raise DuplicateIdError: if layer.ID is identical to a Layer already
                present in the Passage.
            FrozenPassageError: if the :class:`Passage` object we are part of
                is frozen and can't be modified.

        """
        if layer.ID in self._layers:
            raise DuplicateIdError(layer.ID)
        self._layers[layer.ID] = layer

    @ModifyPassage
    def _update_categories(self, category):
        self._categories[category.tag] = {"layer": category.layer, "slot": category.slot, "parent": category.parent}

    @ModifyPassage
    def _update_refined_categories(self, refined_category):
        self._refined_categories.append(refined_category)

    @ModifyPassage
    def _add_node(self, node):
        """Adds a :class:`Node` object to the :class:`Passage`.

        :param node: the Node object to add

        :raise DuplicateIdError: if node.ID is identical to a Node already
                present in the Passage.
            FrozenPassageError: if the :class:`Passage` object we are part of
                is frozen and can't be modified.

        """
        if node.ID in self._nodes:
            raise DuplicateIdError(node.ID)
        self._nodes[node.ID] = node

    def _remove_node(self, node):
        """Removes a :class:`Node` object from the :class:`Passage`.

        :param node: the Node object to remove, must be unlinked with any other
                Node objects and removed from its :class:`Layer`.

        :raise KeyError: if no Node with this ID is present

        """
        del self._nodes[node.ID]

    @ModifyPassage
    def _add_edge(self, edge):
        """Adds a :class:`Edge` object to :class:`Passage`.

        Handles altering the Passage and :class:`Layer` objects accordingly.

        :param edge: the Edge object to add

        """
        # Currently no work is done in the Passage level
        edge.parent.layer._add_edge(edge)

    def _remove_edge(self, edge):
        """Removes a :class:`Edge` object from :class:`Passage`.

        Handles altering the Passage and :class:`Layer` objects accordingly.

        :param edge: the Edge object to remove

        """
        # Currently no work is done in the Passage level
        edge.parent.layer._remove_edge(edge)

    def _change_edge_tag(self, edge, old_tag):
        """Updates the :class:`Passage` and :class:`Layer` objects with the change.

        :param edge: the updated :class:`Edge` object
            old_tag: the Edge's tag before the change

        """
        # Currently no work is done in the Passage level
        edge.parent.layer._change_edge_tag(edge, old_tag)

    def _change_node_tag(self, node, old_tag):
        """Updates the :class:`Passage` and :class:`Layer` objects with the change.

        :param node: the updated :class:`Node` object
            old_tag: the Node's tag before the change

        """
        # Currently no work is done in the Passage level
        node.layer._change_node_tag(node, old_tag)

    def __str__(self):
        try:
            return str(self._layers[max(self._layers)].heads[0])
        except (KeyError, ValueError, IndexError):
            return super().__str__()



import argparse
import sys

from tqdm import tqdm

from ucca.evaluation import evaluate, Scores, LABELED, UNLABELED
from uccaapp.download_task import TaskDownloader

desc = """Download tasks from UCCA-App and evaluate them"""


def main(task_ids, by_filename=False, validate=None, log=None, **kwargs):
    kwargs["write"] = False
    if by_filename:
        task_ids_from_file = []
        for filename in task_ids:
            with open(filename, 'r') as f:
                task_ids_from_file += zip(*list(map(str.split, filter(None, map(str.strip, f)))))
        task_ids = task_ids_from_file
    else:
        task_ids = [[task_id] for task_id in task_ids]
    assert len(task_ids) == 2, "Got %d lists of task IDs instead of two" % len(task_ids)
    downloader = TaskDownloader(**kwargs)
    scores = []
    validate_h = open(validate, "w", encoding="utf-8") if validate else None
    log_h = open(log, "w", encoding="utf-8") if log else None
    if log:
        fields = ["guessed", "ref"] + Scores.field_titles(eval_type=LABELED) + Scores.field_titles(eval_type=UNLABELED)
        print(*fields, file=log_h, sep="\t", flush=True)
    for task_id_pair in tqdm(list(zip(*task_ids)), unit=" tasks", desc="Evaluating"):
        passage_pair = []
        for task_id in task_id_pair:
            passage, *_ = downloader.download_task(task_id, validate=validate_h, **kwargs)
            passage_pair.append(passage)
        score = evaluate(*passage_pair, **kwargs)
        if log:
            fields = list(task_id_pair) + score.fields(eval_type=LABELED) + score.fields(eval_type=UNLABELED)
            print(*fields, file=log_h, sep="\t", flush=True)
        scores.append(score)
    if validate:
        validate_h.close()
    if log:
        log_h.close()
    print()
    if len(scores) > 1:
        print("Aggregated scores:")
    Scores.aggregate(scores).print()


def check_args(p, args):
    if len(args.task_ids) not in (1, 2):
        p.error("Must supply exactly two task IDs or files with IDs, but got %d arguments" % len(args.task_ids))
    return args


if __name__ == "__main__":
    argument_parser = argparse.ArgumentParser(description=desc)
    TaskDownloader.add_arguments(argument_parser)
    main(**vars(check_args(argument_parser, argument_parser.parse_args())))
    sys.exit(0)



import os
from functools import partial
from io import StringIO
from itertools import repeat

import pytest

from ucca import core, layer0, layer1, convert
from ucca.evaluation import evaluate, LABELED, UNLABELED, WEAK_LABELED
from ucca.validation import validate
from .conftest import PASSAGES, load_xml

PRIMARY = "primary"
REMOTE = "remote"
IMPLICIT='implicit'

"""Tests the evaluation module functions and classes."""


def passage1():
    p = core.Passage("1")
    l0 = layer0.Layer0(p)
    l1 = layer1.Layer1(p)
    # 20 terminals (1-20), #10 and #20 are punctuation
    terms = [l0.add_terminal(text=str(i), punct=(i % 10 == 0)) for i in range(1, 21)]

    # Linker #1 with terminal 1
    link1 = l1.add_fnode(None, layer1.EdgeTags.Linker)
    link1.add(layer1.EdgeTags.Terminal, terms[0])

    # Scene #1: [[2 3 4 5 P] [6 7 8 9 A] [10 U] H]
    ps1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p1 = l1.add_fnode(ps1, layer1.EdgeTags.Process)
    a1 = l1.add_fnode(ps1, layer1.EdgeTags.Participant)
    p1.add(layer1.EdgeTags.Terminal, terms[1])
    p1.add(layer1.EdgeTags.Terminal, terms[2])
    p1.add(layer1.EdgeTags.Terminal, terms[3])
    p1.add(layer1.EdgeTags.Terminal, terms[4])
    a1.add(layer1.EdgeTags.Terminal, terms[5])
    a1.add(layer1.EdgeTags.Terminal, terms[6])
    a1.add(layer1.EdgeTags.Terminal, terms[7])
    a1.add(layer1.EdgeTags.Terminal, terms[8])
    l1.add_punct(ps1, terms[9])

    # Scene #23: [[11 12 13 14 15 H] [16 L] [17 18 19 H] H]
    # Scene #2: [[11 12 13 14 A] [15 D]]
    ps23 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    ps2 = l1.add_fnode(ps23, layer1.EdgeTags.ParallelScene)
    a2 = l1.add_fnode(ps2, layer1.EdgeTags.Participant)
    a2.add(layer1.EdgeTags.Terminal, terms[10])
    a2.add(layer1.EdgeTags.Terminal, terms[11])
    a2.add(layer1.EdgeTags.Terminal, terms[12])
    a2.add(layer1.EdgeTags.Terminal, terms[13])
    d2 = l1.add_fnode(ps2, layer1.EdgeTags.Adverbial)
    d2.add(layer1.EdgeTags.Terminal, terms[14])

    # Linker #2: [16 L]
    link2 = l1.add_fnode(ps23, layer1.EdgeTags.Linker)
    link2.add(layer1.EdgeTags.Terminal, terms[15])

    # Scene #3: [[16 17 S] [18 A] (implicit participant) H]
    ps3 = l1.add_fnode(ps23, layer1.EdgeTags.ParallelScene)
    p3 = l1.add_fnode(ps3, layer1.EdgeTags.State)
    p3.add(layer1.EdgeTags.Terminal, terms[16])
    p3.add(layer1.EdgeTags.Terminal, terms[17])
    a3 = l1.add_fnode(ps3, layer1.EdgeTags.Participant)
    a3.add(layer1.EdgeTags.Terminal, terms[18])
    l1.add_fnode(ps3, layer1.EdgeTags.Participant, implicit=True)

    # Punctuation #20 - not under a scene
    l1.add_punct(None, terms[19])

    # adding remote argument to scene #1, remote process to scene #2
    # creating linkages L1->H1, H2<-L2->H3
    l1.add_remote(ps1, layer1.EdgeTags.Participant, d2)
    l1.add_remote(ps2, layer1.EdgeTags.Process, p1)
    l1.add_linkage(link1, ps1)
    l1.add_linkage(link2, ps2, ps3)

    return p


def passage2():
    p = core.Passage("2")
    l0 = layer0.Layer0(p)
    l1 = layer1.Layer1(p)
    # 20 terminals (1-20), #10 and #20 are punctuation
    terms = [l0.add_terminal(text=str(i), punct=(i % 10 == 0)) for i in range(1, 21)]

    # Linker #1 with terminal 1
    link1 = l1.add_fnode(None, layer1.EdgeTags.Linker)  # true
    link1.add(layer1.EdgeTags.Terminal, terms[0])

    # Scene #1: [[2 3 4 5 P] [6 7 8 9 A] [10 U] H]
    ps1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)  # true
    p1 = l1.add_fnode(ps1, layer1.EdgeTags.Process)  # true
    a1 = l1.add_fnode(ps1, layer1.EdgeTags.Participant)  # true
    p1.add(layer1.EdgeTags.Terminal, terms[1])
    p1.add(layer1.EdgeTags.Terminal, terms[2])
    p1.add(layer1.EdgeTags.Terminal, terms[3])
    p1.add(layer1.EdgeTags.Terminal, terms[4])
    a1.add(layer1.EdgeTags.Terminal, terms[5])
    a1.add(layer1.EdgeTags.Terminal, terms[6])
    a1.add(layer1.EdgeTags.Terminal, terms[7])
    a1.add(layer1.EdgeTags.Terminal, terms[8])
    l1.add_punct(ps1, terms[9])

    # Scene #23: [[11 12 13 14 15 H] [16 L] [17 18 19 H] H]
    # Scene #2: [[11 12 13 14 H] [15 E]]
    ps23 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)  # true
    ps2 = l1.add_fnode(ps23, layer1.EdgeTags.ParallelScene)  # true
    a2 = l1.add_fnode(ps2, layer1.EdgeTags.ParallelScene)  # false
    a2.add(layer1.EdgeTags.Terminal, terms[10])
    a2.add(layer1.EdgeTags.Terminal, terms[11])
    a2.add(layer1.EdgeTags.Terminal, terms[12])
    a2.add(layer1.EdgeTags.Terminal, terms[13])
    d2 = l1.add_fnode(ps1, layer1.EdgeTags.Elaborator)  # false
    d2.add(layer1.EdgeTags.Terminal, terms[14])

    # Linker #2: [16 L]
    link2 = l1.add_fnode(ps23, layer1.EdgeTags.Linker)  # true
    link2.add(layer1.EdgeTags.Terminal, terms[15])

    # Scene #3: [[16 17 P] [18 A] (implicit participant) H]
    ps3 = l1.add_fnode(ps23, layer1.EdgeTags.ParallelScene)  # true
    p3 = l1.add_fnode(ps3, layer1.EdgeTags.Process)  # false
    p3.add(layer1.EdgeTags.Terminal, terms[16])
    p3.add(layer1.EdgeTags.Terminal, terms[17])
    a3 = l1.add_fnode(ps3, layer1.EdgeTags.Participant)  # true
    a3.add(layer1.EdgeTags.Terminal, terms[18])
    l1.add_fnode(ps3, layer1.EdgeTags.Participant, implicit=True)

    # Punctuation #20 - not under a scene
    l1.add_punct(None, terms[19])

    # adding remote argument to scene #1, remote process to scene #2
    # creating linkages L1->H1, H2<-L2->H3
    l1.add_remote(ps1, layer1.EdgeTags.Participant, d2)
    l1.add_remote(ps1, layer1.EdgeTags.Participant, a3)
    l1.add_remote(ps2, layer1.EdgeTags.State, p1)
    l1.add_linkage(link1, ps1)
    l1.add_linkage(link2, ps2, ps3)

    return p


def simple1():
    p = core.Passage("1")
    l0 = layer0.Layer0(p)
    l1 = layer1.Layer1(p)
    # 5 terminals (1-5), #5 is punctuation
    terms = [l0.add_terminal(text=str(i), punct=(i == 5)) for i in range(1, 6)]

    # Scene #1: [H [P 1] [A 2]]
    ps1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p1 = l1.add_fnode(ps1, layer1.EdgeTags.Process)
    a = l1.add_fnode(ps1, layer1.EdgeTags.Participant)
    p1.add(layer1.EdgeTags.Terminal, terms[0])
    a.add(layer1.EdgeTags.Terminal, terms[1])

    # Linker #1 with terminal 3
    link1 = l1.add_fnode(None, layer1.EdgeTags.Linker)
    link1.add(layer1.EdgeTags.Terminal, terms[2])

    # Scene #2: [H [A* 2] [S 4]]
    ps2 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p2 = l1.add_fnode(ps2, layer1.EdgeTags.State)
    p2.add(layer1.EdgeTags.Terminal, terms[3])
    l1.add_fnode(ps2, layer1.EdgeTags.Participant, implicit=True)  # implicit should not affect evaluation

    # Punctuation #5 - not under a scene
    l1.add_punct(ps2, terms[4])  # punctuation should not affect evaluation

    # adding remote argument to scene #2
    l1.add_remote(ps2, layer1.EdgeTags.Participant, a)
    l1.add_linkage(link1, ps1, ps2)  # linkage should not affect evaluation

    return p


def simple2():
    p = core.Passage("2")
    l0 = layer0.Layer0(p)
    l1 = layer1.Layer1(p)
    # 5 terminals (1-5), #5 is punctuation
    terms = [l0.add_terminal(text=str(i), punct=(i == 5)) for i in range(1, 6)]

    # Scene #1: [H [S 1] [D 2]]
    ps1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p1 = l1.add_fnode(ps1, layer1.EdgeTags.State)
    a = l1.add_fnode(ps1, layer1.EdgeTags.Adverbial)
    p1.add(layer1.EdgeTags.Terminal, terms[0])
    a.add(layer1.EdgeTags.Terminal, terms[1])

    # Linker #1 with terminal 3
    link1 = l1.add_fnode(None, layer1.EdgeTags.Linker)
    link1.add(layer1.EdgeTags.Terminal, terms[2])

    # Scene #2: [H [A* 2] [S 4]]
    ps2 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p2 = l1.add_fnode(ps2, layer1.EdgeTags.State)
    p2.add(layer1.EdgeTags.Terminal, terms[3])

    # Punctuation #5 - not under a scene
    l1.add_punct(None, terms[4])

    # adding remote argument to scene #2
    l1.add_remote(ps2, layer1.EdgeTags.Adverbial, a)

    return p


def function1():
    p = core.Passage("1")
    l0 = layer0.Layer0(p)
    l1 = layer1.Layer1(p)
    # 5 terminals (1-5), #5 is punctuation
    terms = [l0.add_terminal(text=str(i), punct=(i == 5)) for i in range(1, 6)]

    # Scene #1: [H [P 1] [A 2]]
    ps1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p1 = l1.add_fnode(ps1, layer1.EdgeTags.Process)
    a = l1.add_fnode(ps1, layer1.EdgeTags.Participant)
    p1.add(layer1.EdgeTags.Terminal, terms[0])
    a.add(layer1.EdgeTags.Terminal, terms[1])

    # Function #1 with terminal 3 - its location should not affect evaluation
    f = l1.add_fnode(None, layer1.EdgeTags.Function)
    f.add(layer1.EdgeTags.Terminal, terms[2])

    # Scene #2: [H [A* 2] [S 4]]
    ps2 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p2 = l1.add_fnode(ps2, layer1.EdgeTags.State)
    p2.add(layer1.EdgeTags.Terminal, terms[3])
    l1.add_fnode(ps2, layer1.EdgeTags.Participant, implicit=True)  # implicit should not affect evaluation

    # Punctuation #5 - not under a scene
    l1.add_punct(ps2, terms[4])  # punctuation should not affect evaluation

    # adding remote argument to scene #2
    l1.add_remote(ps2, layer1.EdgeTags.Participant, a)

    return p


def function2():
    p = core.Passage("2")
    l0 = layer0.Layer0(p)
    l1 = layer1.Layer1(p)
    # 5 terminals (1-5), #5 is punctuation
    terms = [l0.add_terminal(text=str(i), punct=(i == 5)) for i in range(1, 6)]

    # Scene #1: [H [S 1] [D 2] [F 2]]
    ps1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p1 = l1.add_fnode(ps1, layer1.EdgeTags.State)
    a = l1.add_fnode(ps1, layer1.EdgeTags.Adverbial)
    p1.add(layer1.EdgeTags.Terminal, terms[0])
    a.add(layer1.EdgeTags.Terminal, terms[1])
    f = l1.add_fnode(ps1, layer1.EdgeTags.Function)
    f.add(layer1.EdgeTags.Terminal, terms[2])

    # Scene #2: [H [A* 2] [S 4]]
    ps2 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p2 = l1.add_fnode(ps2, layer1.EdgeTags.State)
    p2.add(layer1.EdgeTags.Terminal, terms[3])

    # Punctuation #5 - not under a scene
    l1.add_punct(None, terms[4])

    # adding remote argument to scene #2
    l1.add_remote(ps2, layer1.EdgeTags.Adverbial, a)

    return p


def check_primary_remote(scores, expected):
    for (labeled, construction), score in sorted(expected.items()) if hasattr(expected, "items") else \
            zip([(l, c) for l, e in sorted(scores.evaluators.items()) for c in e.results], repeat(expected)):
        buf = StringIO()
        scores.print(file=buf)
        assert score == pytest.approx(scores[labeled][construction].f1), "%s_%s_f1\n%s" % (construction, labeled,
                                                                                           buf.getvalue())


@pytest.mark.parametrize("create", PASSAGES + (passage1, passage2, simple1, simple2, function1, function2))
@pytest.mark.parametrize("units", (True, False), ids=("units", ""))
@pytest.mark.parametrize("errors", (True, False), ids=("errors", ""))
@pytest.mark.parametrize("normalize", (True, False), ids=("normalize", ""))
def test_evaluate_self(create, units, errors, normalize):
    p = create()
    scores = evaluate(p, p, units=units, errors=errors, normalize=normalize)
    assert 1.0 == scores.average_f1()
    for eval_type, results in sorted(scores.evaluators.items()):
        for construction, stats in results.results.items():
            assert 1.0 == stats.f1, (eval_type, construction)
            assert 1.0 == stats.p, (eval_type, construction)
            assert 1.0 == stats.r, (eval_type, construction)
    check_primary_remote(scores, 1.0)


@pytest.mark.parametrize("create1, create2, f1", (
                                 (passage1, passage2, {(LABELED, PRIMARY): 0.5, (LABELED, REMOTE): 0.4, (LABELED, IMPLICIT): 1,
                                                       (UNLABELED, PRIMARY): 0.75, (UNLABELED, REMOTE): 0.8, (UNLABELED, IMPLICIT): 1,
                                                       (WEAK_LABELED, PRIMARY): 7/12, (WEAK_LABELED, REMOTE): 0.8, (WEAK_LABELED, IMPLICIT): 1}),
                                 (simple1, simple2, {(LABELED, PRIMARY): 0.6, (LABELED, REMOTE): 0, (LABELED, IMPLICIT): 0,
                                                     (UNLABELED, PRIMARY): 1, (UNLABELED, REMOTE): 1, (UNLABELED, IMPLICIT): 0,
                                                     (WEAK_LABELED, PRIMARY): 0.8, (WEAK_LABELED, REMOTE): 0, (WEAK_LABELED, IMPLICIT): 0}),
                                 (function1, function2, {(LABELED, PRIMARY): 0.6, (LABELED, REMOTE): 0, (LABELED, IMPLICIT): 0,
                                                         (UNLABELED, PRIMARY): 1, (UNLABELED, REMOTE): 1, (UNLABELED, IMPLICIT): 0,
                                                         (WEAK_LABELED, PRIMARY): 0.8, (WEAK_LABELED, REMOTE): 0, (WEAK_LABELED, IMPLICIT): 0}),
                                 tuple(partial(convert.from_standard, load_xml(
                                     os.path.join("test_files", "%s.xml" % f))) for f in ("120_parsed", "standard3")) +
                                 ({(LABELED, PRIMARY): 3/14, (LABELED, REMOTE): 0, (LABELED, IMPLICIT): 0,
                                   (UNLABELED, PRIMARY): 0.5, (UNLABELED, REMOTE): 0, (UNLABELED, IMPLICIT): 0,
                                   (WEAK_LABELED, PRIMARY): 3/14, (WEAK_LABELED, REMOTE): 0, (WEAK_LABELED, IMPLICIT): 0},),
                                 tuple(partial(convert.from_standard, load_xml(
                                     os.path.join("test_files", "%s.xml" % f))) for f in ("implicit1", "implicit1_ref")) +
                                 ({(LABELED, PRIMARY): 1, (LABELED, REMOTE): 1, (LABELED, IMPLICIT): 2/3,
                                   (UNLABELED, PRIMARY): 1, (UNLABELED, REMOTE): 1, (UNLABELED, IMPLICIT): 2/3,
                                   (WEAK_LABELED, PRIMARY): 1, (WEAK_LABELED, REMOTE): 1, (WEAK_LABELED, IMPLICIT): 2/3},),
                                 tuple(partial(convert.from_standard, load_xml(
                                     os.path.join("test_files", "%s.xml" % f))) for f in ("implicit2", "implicit2_ref")) +
                                 ({(LABELED, PRIMARY): 2/(11/5+13/5), (LABELED, REMOTE): 0, (LABELED, IMPLICIT): 0,
                                   (UNLABELED, PRIMARY): 2/(1+13/11), (UNLABELED, REMOTE): 0, (UNLABELED, IMPLICIT): 0,
                                   (WEAK_LABELED, PRIMARY): 0.5, (WEAK_LABELED, REMOTE): 0, (WEAK_LABELED, IMPLICIT): 0},),
                         ))
def test_evaluate(create1, create2, f1, units=False, errors=False):
    p1 = create1()
    p2 = create2()
    validation_errors_before = [list(validate(p, linkage=False)) for p in (p1, p2)]
    scores = evaluate(p1, p2, units=units, errors=errors, verbose=True)
    validation_errors_after = [list(validate(p, linkage=False)) for p in (p1, p2)]
    for before, after in zip(validation_errors_before, validation_errors_after):
        if not before:
            assert not after
    check_primary_remote(scores, f1)



import os
from argparse import ArgumentParser
from xml.etree.ElementTree import tostring

from tqdm import tqdm

from ucca import convert
from ucca.ioutil import write_passage, external_write_mode
from ucca_db.api import get_by_xids, get_most_recent_passage_by_uid

desc = "Download passages from old UCCA annotation app"


def get_by_method(method, id_field, passage_id=None, **kwargs):
    if method == "xid":
        return get_by_xids(xids=id_field, **kwargs)[0]
    elif method == "uid":
        return get_most_recent_passage_by_uid(id_field, passage_id, **kwargs)
    raise ValueError("Unknown method: '%s'" % method)


def main(args):
    os.makedirs(args.outdir, exist_ok=True)
    with open(args.filename, encoding="utf-8") as f:
        t = list(map(str.split, f))
        if not args.verbose:
            t = tqdm(t, desc="Downloading", unit=" passages")
        for passage_id, id_field in t:
            if not args.verbose:
                t.set_postfix({"passage_id": passage_id, args.method: id_field})
            if args.verbose:
                with external_write_mode():
                    print("Getting passage " + passage_id + " with " + args.method + "=" + id_field, end="\t")
            xml_root = get_by_method(id_field=id_field.split(","), passage_id=passage_id, **vars(args))
            if xml_root is None:
                continue
            if args.write_site:
                site_filename = passage_id + "_site_download.xml"
                with open(site_filename, "w", encoding="utf-8") as fsite:
                    print(tostring(xml_root).decode(), file=fsite)
                if args.verbose:
                    with external_write_mode():
                        print("Wrote '%s'" % site_filename)
            if args.write:
                write_passage(convert.from_site(xml_root), outdir=args.outdir, verbose=args.verbose)


if __name__ == "__main__":
    argparser = ArgumentParser(description=desc)
    argparser.add_argument("filename", help="specification filename with (passage ID, xid OR uid) per passage")
    argparser.add_argument("-m", "--method", default="uid", choices=("xid", "uid"), help="by xid or latest by paid,uid")
    argparser.add_argument("-d", "--db-name", default="work", help="database name")
    argparser.add_argument("-H", "--host-name", default="pgserver", help="host name")
    argparser.add_argument("-o", "--outdir", default=".", help="directory to write created XML IDs to")
    argparser.add_argument("-s", "--write-site", action="store_true", help="write site format, too, for debugging")
    argparser.add_argument("-n", "--no-write", dest="write", action="store_false", help="do not really write any files")
    argparser.add_argument("-x", "--write-xids", help="file to write xids to (for `uid' method)")
    argparser.add_argument("-S", "--strict", action="store_true", help="fail if no result is found")
    argparser.add_argument("-v", "--verbose", action="store_true", help="print tagged text for each passage")
    main(argparser.parse_args())



"""Utility functions for UCCA package."""
import os
import sys
import time
from collections import OrderedDict
from collections import deque
from contextlib import contextmanager
from enum import Enum
from itertools import groupby, islice
from operator import attrgetter, itemgetter

import numpy as np
from tqdm import tqdm

from ucca import layer0, layer1

MODEL_ENV_VAR = "SPACY_MODEL"  # Determines the default spaCy model to load
DEFAULT_MODEL = {"en": "en_core_web_md", "fr": "fr_core_news_md", "de": "de_core_news_md", "ru": "ru"}

N_THREADS = 4
BATCH_SIZE = 50


class Attr(Enum):
    """Wrapper for spaCy Attr, determining order for saving in layer0.extra per token when as_array=True"""
    ORTH = 0
    LEMMA = 1
    TAG = 2
    POS = 3
    ENT_TYPE = 4
    ENT_IOB = 5
    DEP = 6
    HEAD = 7
    SHAPE = 8
    PREFIX = 9
    SUFFIX = 10

    def __call__(self, value, vocab=None, as_array=False, lang=None):
        """Resolve numeric ID of attribute value to string (if as_array=False) or to int (if as_array=True)"""
        if value is None:
            return None
        if self in (Attr.ENT_IOB, Attr.HEAD):
            return int(np.int64(value))
        if as_array:
            is_str = isinstance(value, str)
            if is_str or self in (Attr.ORTH, Attr.LEMMA):
                try:  # Will find the value even if it's a new string, but that's OK since the hash is deterministic
                    i = get_vocab(vocab, lang).strings[value]
                    if is_str:  # Replace with numeric ID since as_array=True
                        value = i
                except KeyError:
                    value = None
            return value if value is None or isinstance(value, str) else int(value)
        try:
            return get_vocab(vocab, lang)[value].text
        except KeyError:
            return None
    
    @property
    def key(self):
        """String used in `extra' dict of Terminals to store this attribute when as_array=False"""
        return self.name.lower()


def get_nlp(lang="en"):
    """ Load spaCy model for a given language, determined by `models' dict or by MODEL_ENV_VAR """
    instance = nlp.get(lang)
    if instance is None:
        model = models.get(lang)
        if not model:
            models[lang] = model = os.environ.get("_".join((MODEL_ENV_VAR, lang.upper()))) or \
                                   os.environ.get(MODEL_ENV_VAR) or DEFAULT_MODEL.get(lang, "xx")
        started = time.time()
        with external_write_mode():
            print("Loading spaCy model '%s'... " % model, end="", flush=True)
        nlp[lang] = instance = load_spacy_model(model)
        with external_write_mode():
            print("Done (%.3fs)." % (time.time() - started))
        tokenizer[lang] = instance.tokenizer
        import spacy
        instance.tokenizer = lambda words: spacy.tokens.Doc(instance.vocab, words=words)
    return instance


def load_spacy_model(model):
    if model == "ru":
        try:
            from spacy.lang.ru import Russian
            return Russian()
        except OSError as e:
            raise OSError("Failed to get spaCy Russian model. Install it using "
                          "pip install git+https://github.com/aatimofeev/spacy_russian_tokenizer.git") from e
    import spacy
    try:
        return spacy.load(model)
    except OSError:
        spacy.cli.download(model)
        # Workaround from https://github.com/explosion/spaCy/issues/3435#issuecomment-474580269
        from spacy.cli import link
        from spacy.util import get_package_path
        link(model, model, force=True, model_path=get_package_path(model))
        try:
            return spacy.load(model)
        except OSError as e:
            raise OSError("Failed to get spaCy model. Download it manually using "
                          "`python -m spacy download %s`." % model) from e


models = {}  # maps language two-letter code to name of spaCy model
nlp = {}  # maps language two-letter code to actual loaded spaCy model
tokenizer = {}  # maps language two-letter code to tokenizer of spaCy model


def get_tokenizer(tokenized=False, lang="en"):
    instance = get_nlp(lang)
    return instance.tokenizer if tokenized else tokenizer[lang]


def get_vocab(vocab=None, lang=None):
    if vocab is not None:
        return vocab
    return (get_nlp(lang) if lang else get_nlp()).vocab


def get_word_vectors(dim=None, size=None, filename=None, vocab=None):
    """
    Get word vectors from spaCy model or from text file
    :param dim: dimension to trim vectors to (default: keep original)
    :param size: maximum number of vectors to load (default: all)
    :param filename: text file to load vectors from (default: from spaCy model)
    :param vocab: instead of strings, look up keys of returned dict in vocab (use lang str, e.g. "en", for spaCy vocab)
    :return: tuple of (dict of word [string or integer] -> vector [NumPy array], dimension)
    """
    orig_keys = vocab is None
    if isinstance(vocab, str) or not filename:
        vocab = get_nlp(vocab if isinstance(vocab, str) else "en").vocab

    def _lookup(word):
        try:
            return word.orth_ if orig_keys else word.orth
        except AttributeError:
            if orig_keys:
                return word
        lex = vocab[word]
        return getattr(lex, "orth", lex)

    if filename:
        it = read_word_vectors(dim, size, filename)
        nr_row, nr_dim = next(it)
        vectors = OrderedDict(islice(tqdm(((_lookup(w), v) for w, v in it if orig_keys or w in vocab),
                                          desc="Loading '%s'" % filename, postfix=dict(dim=nr_dim),
                                          file=sys.stdout, total=nr_row, unit=" vectors"), nr_row))
    else:  # return spaCy vectors
        nr_row, nr_dim = vocab.vectors.shape
        if dim is not None and dim < nr_dim:
            nr_dim = int(dim)
            vocab.vectors.resize(shape=(int(size or nr_row), nr_dim))
        lexemes = sorted([l for l in vocab if l.has_vector], key=attrgetter("prob"), reverse=True)[:size]
        vectors = OrderedDict((_lookup(l), l.vector) for l in lexemes)
    return vectors, nr_dim


def read_word_vectors(dim, size, filename):
    """
    Read word vectors from text file, with an optional first row indicating size and dimension
    :param dim: dimension to trim vectors to
    :param size: maximum number of vectors to load
    :param filename: text file to load vectors from
    :return: generator: first element is (#vectors, #dims); and all the rest are (word [string], vector [NumPy array])
    """
    try:
        first_line = True
        nr_row = nr_dim = None
        with open(filename, encoding="utf-8") as f:
            for line in f:
                fields = line.split()
                if first_line:
                    first_line = False
                    try:
                        nr_row, nr_dim = map(int, fields)
                        is_header = True
                    except ValueError:
                        nr_dim = len(fields) - 1  # No header, just get vector length from first one
                        is_header = False
                    if dim and dim < nr_dim:
                        nr_dim = dim
                    yield size or nr_row, nr_dim
                    if is_header:
                        continue  # Read next line
                word, *vector = fields
                if len(vector) >= nr_dim:  # May not be equal if word is whitespace
                    yield word, np.asarray(vector[-nr_dim:], dtype="f")
    except OSError as e:
        raise IOError("Failed loading word vectors from '%s'" % filename) from e


def annotate(passage, *args, **kwargs):
    """
    Run spaCy pipeline on the given passage, unless already annotated
    :param passage: Passage object, whose layer 0 nodes will be added entries in the `extra' dict
    """
    list(annotate_all([passage], *args, **kwargs))


def annotate_as_tuples(passages, replace=False, as_array=False, as_extra=True, lang="en", vocab=None, verbose=False):
    for passage_lang, passages_by_lang in groupby(passages, get_lang):
        for need_annotation, stream in groupby(to_annotate(passages_by_lang, replace, as_array, as_extra),
                                               lambda x: bool(x[0])):
            annotated = get_nlp(passage_lang or lang).pipe(
                stream, as_tuples=True, n_threads=N_THREADS, batch_size=BATCH_SIZE) if need_annotation else stream
            annotated = set_docs(annotated, as_array, as_extra, passage_lang or lang, vocab, replace, verbose)
            for passage, passages in groupby(annotated, itemgetter(0)):
                yield deque(passages, maxlen=1).pop()  # Wait until all paragraphs have been annotated


def annotate_all(passages, replace=False, as_array=False, as_extra=True, as_tuples=False, lang="en", vocab=None,
                 verbose=False):
    """
    Run spaCy pipeline on the given passages, unless already annotated
    :param passages: iterable of Passage objects, whose layer 0 nodes will be added entries in the `extra' dict
    :param replace: even if a given passage is already annotated, replace with new annotation
    :param as_array: instead of adding `extra' entries to each terminal, set layer 0 extra["doc"] to array of ids
    :param as_extra: set `extra' entries to each terminal
    :param as_tuples: treat input as tuples of (passage text, context), and return context for each passage as-is
    :param lang: optional two-letter language code, will be overridden if passage has "lang" attrib
    :param vocab: optional dictionary of vocabulary IDs to string values, to avoid loading spaCy model
    :param verbose: whether to print annotated text
    :return: generator of annotated passages, which are actually modified in-place (same objects as input)
    """
    if not as_tuples:
        passages = ((p,) for p in passages)
    for t in annotate_as_tuples(passages, replace=replace, as_array=as_array, as_extra=as_extra, lang=lang, vocab=vocab,
                                verbose=verbose):
        yield t if as_tuples else t[0]


def get_lang(passage_context):
    return passage_context[0].attrib.get("lang")


def to_annotate(passage_contexts, replace, as_array=False, as_extra=True):
    """Filter passages to get only those that require annotation; split to paragraphs and return generator of
    (list of tokens, (paragraph index, list of Terminals, Passage) + original context appended) tuples"""
    return (([t.text for t in terminals] if replace or not is_annotated(passage, as_array, as_extra) else (),
             (i, terminals, passage) + tuple(context)) for passage, *context in passage_contexts
            for i, terminals in enumerate(break2paragraphs(passage, return_terminals=True)))


def is_annotated(passage, as_array=False, as_extra=True):
    """Whether the passage is already annotated or only partially annotated"""
    l0 = passage.layer(layer0.LAYER_ID)
    docs = l0.extra.get("doc")
    if as_array:
        if not (not l0.all or docs is not None and len(docs) == max(t.paragraph for t in l0.all) and
                sum(map(len, docs)) == len(l0.all) and
                all(i is None or isinstance(i, int) for l in docs for t in l for i in t)):
            return False
    if as_extra:
        if not all(a.key in t.extra for t in l0.all for a in Attr):
            return False
    return True


def set_docs(annotated, as_array, as_extra, lang, vocab, replace, verbose):
    """Given spaCy annotations, set values in layer0.extra per paragraph if as_array=True,
       and in Terminal.extra if as_extra=True"""
    for doc, (i, terminals, passage, *context) in annotated:
        if doc:  # Not empty, so copy values
            from spacy import attrs
            arr = doc.to_array([getattr(attrs, a.name) for a in Attr])
            if as_array:
                docs = passage.layer(layer0.LAYER_ID).docs(i + 1)
                existing = docs[i] + (len(arr) - len(docs[i])) * [len(Attr) * [None]]
                docs[i] = [[a(v if e is None or replace else e, get_vocab(vocab, lang), as_array=True)
                            for a, v, e in zip(Attr, values, es)] for values, es in zip(arr, existing)]
            if as_extra:
                for terminal, values in zip(terminals, arr):
                    for attr, value in zip(Attr, values):
                        if replace or not terminal.extra.get(attr.key):
                            terminal.extra[attr.key] = attr(value, get_vocab(vocab, lang))
        if verbose:
            data = [[a.key for a in Attr]] + \
                   [[str(a(t.tok[a.value], get_vocab(vocab, lang)) if as_array else t.extra[a.key])
                     for a in Attr] for j, t in enumerate(terminals)]
            width = [max(len(f) for f in t) for t in data]
            for j in range(len(Attr)):
                try:
                    print(" ".join("%-*s" % (w, f[j]) for f, w in zip(data, width)))
                except UnicodeEncodeError:
                    pass
            print()
        yield (passage,) + tuple(context)


SENTENCE_END_MARKS = ('.', '?', '!')
QUOTES = ("'", '"', "`", "»", "«")


def break2sentences(passage, lang="en", *args, **kwargs):
    """
    Breaks paragraphs into sentences according to the annotation.

    A sentence is a list of terminals which ends with a mark from
    SENTENCE_END_MARKS, and is also the end of a paragraph or parallel scene.
    :param passage: the Passage object to operate on
    :param lang: optional two-letter language code
    :return: a list of positions in the Passage, each denotes a closing Terminal of a sentence.
    """
    del args, kwargs
    l1 = passage.layer(layer1.LAYER_ID)
    terminals = extract_terminals(passage)
    if not terminals:
        return []
    if any(n.outgoing for n in l1.all):  # Passage is labeled
        ps_ends = [ps.end_position for ps in l1.top_scenes]
        ps_starts = [ps.start_position for ps in l1.top_scenes]
        marks = []
        for terminal in terminals:
            # Annotations doesn't always include the ending period (or other mark)
            # with the parallel scene it closes. Hence, if the terminal before the
            # mark closed the parallel scene, and this mark doesn't open a scene
            # in any way (hence it probably just "hangs" there), it's a sentence end
            if terminal.text in SENTENCE_END_MARKS and \
                    (terminal.position in ps_ends or
                     (terminal.position - 1) in ps_ends and terminal.position not in ps_starts) or \
                    terminal.position - 1 in marks and layer0.is_punct(terminal) and not \
                    (terminal.text in QUOTES and terminal.text == terminals[marks[-1] - 1].text):
                marks.append(terminal.position)
    else:  # Not labeled, split using spaCy
        annotated = get_nlp(lang=lang)([t.text for t in terminals])
        marks = [span.end for span in annotated.sents]
    marks = sorted(set(marks + break2paragraphs(passage)))
    # Avoid punctuation-only sentences by picking the last punctuation symbol in each consecutive sequence
    if len(marks) > 1:
        marks = [x for x, y in zip(marks[:-1], marks[1:]) if not all(map(layer0.is_punct, terminals[x - 1:y - 1]))] + \
                [marks[-1]]
    return marks


def extract_terminals(p):
    """returns an iterator of the terminals of the passage p"""
    return p.layer(layer0.LAYER_ID).all


def break2paragraphs(passage, return_terminals=False, *args, **kwargs):
    """
    Breaks into paragraphs according to the annotation.

    Uses the `paragraph' attribute of layer 0 to find paragraphs.
    :param passage: the Passage object to operate on
    :param return_terminals: whether to return actual Terminal objects of all terminals rather than just end positions
    :return: a list of positions in the Passage, each denotes a closing Terminal of a paragraph.
    """
    del args, kwargs
    terminals = sorted(extract_terminals(passage), key=attrgetter("position"))
    if not terminals:
        return []
    return [list(p) for _, p in groupby(terminals, key=attrgetter("paragraph"))] if return_terminals else \
        [t1.position for t1, t2 in zip(terminals[:-1], terminals[1:])
         if t2.para_pos == 1 or t1.paragraph != t2.paragraph] + [terminals[-1].position]


def indent_xml(xml_as_string):
    """
    Indents a string of XML-like objects.

    This works only for units with no text or tail members, and only for
    strings whose leaves are written as <tag /> and not <tag></tag>.
    :param xml_as_string: XML string to indent
    :return: indented XML string
    """
    tabs = 0
    lines = str(xml_as_string).replace('><', '>\n<').splitlines()
    s = ''
    for line in lines:
        if line.startswith('</'):
            tabs -= 1
        s += ("  " * tabs) + line + '\n'
        if not (line.endswith('/>') or line.startswith('</')):
            tabs += 1
    return s


@contextmanager
def external_write_mode(*args, **kwargs):
    try:
        with tqdm.external_write_mode(*args, **kwargs):
            yield
    except AttributeError:
        yield



import pytest

from ucca import layer1
from ucca.validation import validate
from .conftest import loaded, loaded_valid, multi_sent, crossing, discontiguous, l1_passage, empty, \
    create_passage, attach_terminals

"""Tests the validation module functions and classes."""


def unary_punct():
    p, l1, terms = create_passage(3, 3)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p1 = l1.add_fnode(h1, layer1.EdgeTags.Process)
    a1 = l1.add_fnode(h1, layer1.EdgeTags.Participant)
    l1.add_punct(h1, terms[2])
    attach_terminals(terms, p1, a1)
    return p


def binary_punct():
    p, l1, terms = create_passage(4, 3)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p1 = l1.add_fnode(h1, layer1.EdgeTags.Process)
    a1 = l1.add_fnode(h1, layer1.EdgeTags.Participant)
    l1.add_punct(h1, terms[2]).add(layer1.EdgeTags.Terminal, terms[3])
    attach_terminals(terms, p1, a1)
    return p


def unary_punct_under_fn():
    p, l1, terms = create_passage(3, 3)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p1 = l1.add_fnode(h1, layer1.EdgeTags.Process)
    a1 = l1.add_fnode(h1, layer1.EdgeTags.Participant)
    attach_terminals(terms, p1, a1, h1)
    return p


def punct_under_unanalyzable_fn():
    p, l1, terms = create_passage(3, 2)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p1 = l1.add_fnode(h1, layer1.EdgeTags.Process)
    attach_terminals(terms, p1, p1, p1)
    return p

def forbid_child_of_P_fn(): # same for adverbial, linker, time, quantifier and connector
    p, l1, terms = create_passage(5, 5)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p1 = l1.add_fnode(h1, layer1.EdgeTags.Process)
    a1 = l1.add_fnode(h1, layer1.EdgeTags.Participant)
    g1 = l1.add_fnode(p1, layer1.EdgeTags.Ground)
    d1 = l1.add_fnode(p1, layer1.EdgeTags.Adverbial)
    attach_terminals(terms, p1, a1, g1, d1)
    return p

def forbid_child_of_S_fn():
    p, l1, terms = create_passage(4, 4)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    s1 = l1.add_fnode(h1, layer1.EdgeTags.State)
    a1 = l1.add_fnode(h1, layer1.EdgeTags.Participant)
    h2 = l1.add_fnode(s1, layer1.EdgeTags.ParallelScene)
    attach_terminals(terms, s1, a1, h2)
    return p

def forbid_child_of_F_fn():
    p, l1, terms = create_passage(6, 6)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p1 = l1.add_fnode(h1, layer1.EdgeTags.Process)
    f1 = l1.add_fnode(h1, layer1.EdgeTags.Function)
    a1 = l1.add_fnode(h1, layer1.EdgeTags.Participant)
    t1 = l1.add_fnode(f1, layer1.EdgeTags.Time)
    s1 = l1.add_fnode(f1, layer1.EdgeTags.State)
    attach_terminals(terms, f1, a1, p1, t1, s1)
    return p

def forbid_child_of_G_fn():
    p, l1, terms = create_passage(3, 3)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    g1 = l1.add_fnode(h1, layer1.EdgeTags.Ground)
    h2 = l1.add_fnode(g1, layer1.EdgeTags.ParallelScene)
    attach_terminals(terms, g1, h2)
    return p

def forbid_child_of_H_fn():
    p, l1, terms = create_passage(3, 3)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    link1 = l1.add_fnode(h1, layer1.EdgeTags.Linker)
    h2 = l1.add_fnode(h1, layer1.EdgeTags.ParallelScene)
    s1 = l1.add_fnode(h2, layer1.EdgeTags.State)
    a1 = l1.add_fnode(h2, layer1.EdgeTags.Participant)
    attach_terminals(terms, link1, s1, a1)
    return p

def forbid_descendant_of_P_fn(): # same for state
    p, l1, terms = create_passage(5, 5)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p1 = l1.add_fnode(h1, layer1.EdgeTags.Process)
    a1 = l1.add_fnode(h1, layer1.EdgeTags.Participant)
    c1 = l1.add_fnode(p1, layer1.EdgeTags.Center)
    e1 = l1.add_fnode(p1, layer1.EdgeTags.Elaborator)
    h2 = l1.add_fnode(c1, layer1.EdgeTags.ParallelScene)
    c2 = l1.add_fnode(e1, layer1.EdgeTags.Center)
    s2 = l1.add_fnode(c2, layer1.EdgeTags.State)
    a2 = l1.add_fnode(c2, layer1.EdgeTags.Participant)
    attach_terminals(terms, a1, e1, a2,s2,h2, e1)
    return p

def forbid_sibling_of_L_H_fn():
    p, l1, terms = create_passage(6, 6)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p1 = l1.add_fnode(h1, layer1.EdgeTags.Process)
    h2 = l1.add_fnode(h1, layer1.EdgeTags.ParallelScene)
    link1 = l1.add_fnode(h1, layer1.EdgeTags.Linker)
    attach_terminals(terms, p1, h2, link1)
    return p

def forbid_sibling_of_S_fn():
    p, l1, terms = create_passage(6, 6)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    s1 = l1.add_fnode(h1, layer1.EdgeTags.State)
    h2 = l1.add_fnode(h1, layer1.EdgeTags.ParallelScene)
    link1 = l1.add_fnode(h1, layer1.EdgeTags.Linker)
    attach_terminals(terms, s1, h2, link1)
    return p

def forbid_sibling_of_D_fn():
    p, l1, terms = create_passage(6, 6)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    d1 = l1.add_fnode(h1, layer1.EdgeTags.Adverbial)
    s1 = l1.add_fnode(h1, layer1.EdgeTags.State)
    h2 = l1.add_fnode(h1, layer1.EdgeTags.ParallelScene)
    link1 = l1.add_fnode(h1, layer1.EdgeTags.Linker)
    attach_terminals(terms, d1, s1, h2, link1)
    return p

def forbid_sibling_of_P_N_fn():
    p, l1, terms = create_passage(6, 6)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p1 = l1.add_fnode(h1, layer1.EdgeTags.Process)
    n1 = l1.add_fnode(h1, layer1.EdgeTags.Connector)
    attach_terminals(terms, n1, p1)
    return p

def forbid_sibling_of_Q_fn():
    p, l1, terms = create_passage(6, 6)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p1 = l1.add_fnode(h1, layer1.EdgeTags.Process)
    q1 = l1.add_fnode(h1, layer1.EdgeTags.Quantifier)
    attach_terminals(terms, p1, q1)
    return p

def forbid_sibling_of_E_fn():
    p, l1, terms = create_passage(6, 6)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p1 = l1.add_fnode(h1, layer1.EdgeTags.Process)
    e1 = l1.add_fnode(h1, layer1.EdgeTags.Elaborator)
    attach_terminals(terms, p1, e1)
    return p

def forbid_sibling_of_C_fn():
    p, l1, terms = create_passage(6, 6)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p1 = l1.add_fnode(h1, layer1.EdgeTags.Process)
    c1 = l1.add_fnode(h1, layer1.EdgeTags.Center)
    attach_terminals(terms, p1, c1)
    return p

def forbid_sibling_of_A_T_fn():
    p, l1, terms = create_passage(6, 6)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p1 = l1.add_fnode(h1, layer1.EdgeTags.Process)
    a1 = l1.add_fnode(h1, layer1.EdgeTags.Participant)
    t1 = l1.add_fnode(h1, layer1.EdgeTags.Time)
    n1 = l1.add_fnode(h1, layer1.EdgeTags.Connector)
    e1 = l1.add_fnode(h1, layer1.EdgeTags.Elaborator)
    c1 = l1.add_fnode(h1, layer1.EdgeTags.Center)
    attach_terminals(terms, p1, a1, t1, n1, e1, c1)
    return p

def require_sibling_of_E_Q_N():
    p, l1, terms = create_passage(6, 6)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    e1 = l1.add_fnode(h1, layer1.EdgeTags.Elaborator)
    q1 = l1.add_fnode(h1, layer1.EdgeTags.Quantifier)
    n1 = l1.add_fnode(h1, layer1.EdgeTags.Connector)
    attach_terminals(terms, e1, q1, n1)
    return p

def require_sibling_of_L():
    p, l1, terms = create_passage(6, 6)
    link1 = l1.add_fnode(None, layer1.EdgeTags.Linker)
    attach_terminals(terms, link1)
    return p

def require_sibling_of_A_T():
    p, l1, terms = create_passage(6, 6)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    a1 = l1.add_fnode(h1, layer1.EdgeTags.Participant)
    t1 = l1.add_fnode(h1, layer1.EdgeTags.Time)
    attach_terminals(terms, a1, t1)
    return p

def unique_under_parent():
    p, l1, terms = create_passage(6, 6)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p1 = l1.add_fnode(h1, layer1.EdgeTags.Process)
    p2 = l1.add_fnode(h1, layer1.EdgeTags.Process)
    a1 = l1.add_fnode(h1, layer1.EdgeTags.Participant)
    attach_terminals(terms, a1, p1, p2)
    return p

def forbid_remote():
    p, l1, terms = create_passage(6, 6)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    h2 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    link1 = l1.add_fnode(None, layer1.EdgeTags.Linker)
    p1 = l1.add_fnode(h1, layer1.EdgeTags.Process)
    p2 = l1.add_fnode(h2, layer1.EdgeTags.Process)
    a1 = l1.add_fnode(h1, layer1.EdgeTags.Participant)
    a2 = l1.add_fnode(h1, layer1.EdgeTags.Participant)
    r1 = l1.add_fnode(h1, layer1.EdgeTags.Relator)
    r2 = l1.add_remote_multiple(h2, layer1.EdgeTags.Relator, r1)
    attach_terminals(terms, link1, a2, a1, p1, p2, r2, r1)
    return p

def forbid_at_top_level():
    p, l1, terms = create_passage(1, 1)
    g1 = l1.add_fnode(None, layer1.EdgeTags.Ground)
    attach_terminals(terms, g1)
    return p

def forbid_children_of_UNA_and_alone():
    p, l1, terms = create_passage(1, 1)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    u1 = l1.add_fnode(h1, layer1.EdgeTags.Unanalyzable)
    s1 = l1.add_fnode(u1, layer1.EdgeTags.State)
    a1 = l1.add_fnode(u1, layer1.EdgeTags.Participant)
    attach_terminals(terms)
    return p


@pytest.mark.parametrize("create, valid", (
        (loaded, False),
        (loaded_valid, True),
        (multi_sent, True),
        (crossing, True),
        (discontiguous, True),
        (l1_passage, True),
        (empty, False),
        (unary_punct, True),
        (binary_punct, True),
        (unary_punct_under_fn, False),
        (punct_under_unanalyzable_fn, True),
        (forbid_child_of_P_fn, False),
        (forbid_child_of_F_fn, False),
        (forbid_child_of_S_fn, False),
        (forbid_child_of_G_fn, False),
        (forbid_child_of_H_fn, False),
        (forbid_descendant_of_P_fn, False),
        (forbid_sibling_of_L_H_fn, False),
        (forbid_sibling_of_S_fn, False),
        (forbid_sibling_of_D_fn, False),
        (forbid_sibling_of_P_N_fn, False),
        (forbid_sibling_of_Q_fn, False),
        (forbid_sibling_of_E_fn, False),
        (forbid_sibling_of_C_fn, False),
        (forbid_sibling_of_A_T_fn, False),
        (require_sibling_of_E_Q_N, False),
        (require_sibling_of_L, False),
        (require_sibling_of_A_T, False),
        (unique_under_parent, False),
        (forbid_remote, False),
        (forbid_at_top_level, False),
        (forbid_children_of_UNA_and_alone, False),
))
def test_evaluate_self(create, valid):
    p = create()
    errors = list(validate(p))
    if valid:
        assert not errors, p
    else:
        assert errors, p



import os.path
import sys
from argparse import ArgumentParser

from ucca.convert import from_text, to_json
from uccaapp.api import ServerAccessor

desc = """
Read input file as one line per paragraph, where paragraphs are separated by multiple newlines and an optional
<DELIMITER>.
Tokenize and upload as submitted tokenization tasks, then create annotation tasks from them.

Tokenization in Russian requires:
    pip install git+https://github.com/aatimofeev/spacy_russian_tokenizer.git
"""


class TokenizerUploader(ServerAccessor):
    def __init__(self, user_id, source_id, project_id, lang=None, **kwargs):
        super().__init__(**kwargs)
        self.set_source(source_id)
        self.set_project(project_id)
        self.set_user(user_id)

    def tokenize_and_upload(self, filename, log=None, lang=None, **kwargs):
        del kwargs
        log_h = open(log, "w", encoding="utf-8") if log else None
        prefix = os.path.splitext(os.path.basename(filename))[0].replace(" ", "_")
        with open(filename, encoding="utf-8") as f:
            for passage, text in from_text(f, passage_id=prefix, lang=lang, return_text=True):
                passage_out = self.create_passage(text=text, type="PUBLIC", source=self.source)
                task_in = dict(type="TOKENIZATION", status="SUBMITTED", project=self.project,
                               user=self.user, passage=passage_out, manager_comment=passage.ID,
                               user_comment="", parent=None, is_demo=False, is_active=True)
                tok_task_out = self.create_task(**task_in)
                tok_user_task_in = dict(tok_task_out)
                tok_user_task_in.update(to_json(passage, return_dict=True, tok_task=True))
                self.submit_task(**tok_user_task_in)
                task_in.update(parent=tok_task_out, type="ANNOTATION")
                ann_user_task_out = self.create_task(**task_in)
                print("Uploaded passage " + filename + " successfully.", file=sys.stderr)
                if log:
                    print(passage.ID, passage_out["id"], tok_task_out["id"], ann_user_task_out["id"],
                          file=log_h, sep="\t", flush=True)
        if log:
            log_h.close()

    @staticmethod
    def add_arguments(argparser):
        argparser.add_argument("filename", help="text file with one line paragraph, where paragraphs are separated "
                                                "by multiple newlines and an optional <DELIMITER>")
        argparser.add_argument("-l", "--log", help="filename to write log of uploaded passages to")
        argparser.add_argument("--lang", choices=["ru", "en", "fr", "de"], default="ru",
                               help="language two-letter code, for tokenizer")
        ServerAccessor.add_project_id_argument(argparser)
        ServerAccessor.add_source_id_argument(argparser)
        ServerAccessor.add_user_id_argument(argparser)
        ServerAccessor.add_arguments(argparser)


def main(**kwargs):
    TokenizerUploader(**kwargs).tokenize_and_upload(**kwargs)


if __name__ == "__main__":
    argument_parser = ArgumentParser(description=desc)
    TokenizerUploader.add_arguments(argument_parser)
    main(**vars(argument_parser.parse_args()))



from ucca import layer1
from .conftest import l1_passage, discontiguous

"""Tests layer1 module functionality and correctness."""


def test_creation():
    p = l1_passage()
    head = p.layer("1").heads[0]
    assert [x.tag for x in head] == ["L", "H", "H", "L", "H", "U"]
    assert [x.child.position for x in head.children[0]] == [1]
    assert [x.tag for x in head.children[1]] == ["P", "A", "U", "A"]
    assert [x.child.position for x in head.children[1].children[0]] == [2, 3, 4, 5]
    assert [x.child.position for x in head.children[1].children[1]] == [6, 7, 8, 9]
    assert [x.child.position for x in head.children[1].children[2]] == [10]
    assert (head.children[1][3].attrib.get("remote"))


def test_fnodes():
    p = l1_passage()
    l0 = p.layer("0")
    l1 = p.layer("1")

    terms = l0.all
    head, lkg1, lkg2 = l1.heads
    link1, ps1, ps2, link2, ps3, punct2 = head.children
    p1, a1, punct1 = [x.child for x in ps1 if not x.attrib.get("remote")]
    a2, d2 = [x.child for x in ps2 if not x.attrib.get("remote")]
    p3, a3, a4 = ps3.children

    assert lkg1.relation == link1
    assert lkg1.arguments == [ps1]
    assert ps2.process == p1
    assert ps1.participants == [a1, d2]
    assert ps3.participants == [a3, a4]

    assert ps1.get_terminals() == terms[1:10]
    assert ps1.get_terminals(punct=False, remotes=True) == terms[1:9] + terms[14:15]
    assert ps1.end_position == 10
    assert ps2.start_position == 11
    assert ps3.start_position == 17
    assert a4.start_position == -1

    assert ps1.fparent == head
    assert d2.fparent == ps2


def test_layer1():
    p = l1_passage()
    l1 = p.layer("1")

    head, lkg1, lkg2 = l1.heads
    link1, ps1, ps2, link2, ps3, punct2 = head.children
    p1, a1, punct1 = [x.child for x in ps1 if not x.attrib.get("remote")]

    assert l1.top_scenes == [ps1, ps2, ps3]
    assert l1.top_linkages == [lkg1, lkg2]

    # Changing the process tag of scene #1 to A and back, validate that
    # top scenes are updates accordingly
    p_edge = [e for e in ps1 if e.tag == layer1.EdgeTags.Process][0]
    p_edge.tag = layer1.EdgeTags.Participant
    assert l1.top_linkages == [lkg2]
    p_edge.tag = layer1.EdgeTags.Process
    assert l1.top_scenes == [ps1, ps2, ps3]
    assert l1.top_linkages == [lkg1, lkg2]


def test_str():
    p = l1_passage()
    assert [str(x) for x in p.layer("1").heads] == \
           ["[L 1] [H [P 2 3 4 5] [A 6 7 8 9] [U 10] "
            "... [A* 15] ] [H [P* 2 3 4 5] [A 11 12 "
            "13 14] [D 15] ] [L 16] [H [A IMPLICIT] [S "
            "17 18] [A 19] ] [U 20] ",
            "1.2-->1.3", "1.10-->1.7,1.11"]


def test_destroy():
    p = l1_passage()
    l1 = p.layer("1")

    head, lkg1, lkg2 = l1.heads
    link1, ps1, ps2, link2, ps3, punct2 = head.children
    p1, a1, punct1 = [x.child for x in ps1 if not x.attrib.get("remote")]

    ps1.destroy()
    assert head.children == [link1, ps2, link2, ps3, punct2]
    assert p1.parents == [ps2]
    assert not a1.parents
    assert not punct1.parents


def test_discontiguous():
    """Tests FNode.discontiguous and FNode.get_sequences"""
    p = discontiguous()
    l1 = p.layer("1")
    head = l1.heads[0]
    ps1, ps2, ps3 = head.children
    d1, a1, p1, f1 = ps1.children
    e1, c1, e2 = d1.children
    d2, g2, p2, a2 = ps2.children
    t14, p3, a3 = ps3.children

    # Checking discontiguous property
    assert not ps1.discontiguous
    assert not d1.discontiguous
    assert not e1.discontiguous
    assert not e2.discontiguous
    assert c1.discontiguous
    assert a1.discontiguous
    assert p1.discontiguous
    assert not f1.discontiguous
    assert ps2.discontiguous
    assert not p2.discontiguous
    assert not a2.discontiguous
    assert not ps3.discontiguous
    assert not a3.discontiguous

    # Checking get_sequences -- should return only non-remote, non-implicit
    # stretches of terminals
    assert ps1.get_sequences() == [(1, 10)]
    assert d1.get_sequences() == [(1, 4)]
    assert e1.get_sequences() == [(1, 1)]
    assert e2.get_sequences() == [(3, 3)]
    assert c1.get_sequences() == [(2, 2), (4, 4)]
    assert a1.get_sequences() == [(5, 5), (8, 8)]
    assert p1.get_sequences() == [(6, 7), (10, 10)]
    assert f1.get_sequences() == [(9, 9)]
    assert ps2.get_sequences() == [(11, 14), (18, 20)]
    assert p2.get_sequences() == [(11, 14)]
    assert a2.get_sequences() == [(18, 20)]
    assert not d2.get_sequences()
    assert not g2.get_sequences()
    assert ps3.get_sequences() == [(15, 17)]
    assert a3.get_sequences() == [(16, 17)]
    assert not p3.get_sequences()



#!/usr/bin/env python3

import argparse
from collections import Counter

from ucca import layer1
from ucca.ioutil import get_passages_with_progress_bar

desc = """Finds edge tags that are empirically always unique: occur at most once in edges per node
"""


def main(args):
    out = args.direction == "out"
    roles = set(tag for name, tag in layer1.EdgeTags.__dict__.items()
                if isinstance(tag, str) and not name.startswith('__'))
    for passage in get_passages_with_progress_bar([args.directory]):
        for node in passage.layer(layer1.LAYER_ID).all:
            counts = Counter(edge.tag for edge in (node if out else node.incoming))
            roles.difference_update(tag for tag, count in counts.items() if count > 1)

    lines = "\n".join(sorted(roles))
    print(lines)
    if args.outfile:
        with open(args.outfile, "w", encoding="utf-8") as f:
            print(lines, file=f)


if __name__ == '__main__':
    argparser = argparse.ArgumentParser(description=desc)
    argparser.add_argument('-d', '--directory', required=True, help="directory with passage files to process")
    argparser.add_argument('-o', '--outfile', default="data/unique_roles.txt", help="output file for data")
    argparser.add_argument('-D', '--direction', default="out", help="direction of edges to check (out|in)")
    main(argparser.parse_args())



import argparse
import string
from glob import glob

from tqdm import tqdm

from ucca import core, layer0, layer1
from ucca.ioutil import write_passage

PUNCTUATION = set(string.punctuation)


def gen_lines(patterns):
    for pattern in patterns:
        for filename in glob(pattern) or [pattern]:
            with open(filename, encoding="utf-8") as f:
                for line in f:
                    line = line.strip()
                    if line:
                        yield line


def main(args):
    for i, line in enumerate(tqdm(gen_lines(args.filenames), unit=" lines", desc="Creating passages"), start=1):
        p = core.Passage(args.format % i)
        l0 = layer0.Layer0(p)
        layer1.Layer1(p)
        for tok in line.split():
            l0.add_terminal(text=tok, punct=PUNCTUATION.issuperset(tok))
        write_passage(p, outdir=args.out_dir, binary=args.binary, verbose=False)


if __name__ == "__main__":
    argparser = argparse.ArgumentParser(description="Create unannotated passage files from tokenized and split text")
    argparser.add_argument("filenames", nargs="+", help="Input filenames containing tokenized and sentence-split text")
    argparser.add_argument("-o", "--out-dir", help="Directory to write output files to")
    argparser.add_argument("-f", "--format", default="1%04d0", help="String format for passage IDs")
    argparser.add_argument("-b", "--binary", action="store_true", help="Write Pickle files instead of XML")
    main(argparser.parse_args())



import csv
import os
from argparse import ArgumentParser

from ucca import layer0, layer1, textutil
from ucca.ioutil import get_passages_with_progress_bar, write_passage
from ucca.normalization import fparent, remove, copy_edge, destroy
from ucca.textutil import annotate_all, Attr

desc = """Convert the English Wiki corpus from version 1.0 to 1.2"""


def get_annotation(terminal, attr):
    return terminal.extra[attr.key]


def is_main_relation(node):
    while node is not None and node.ftag == layer1.EdgeTags.Center:
        node = fparent(node)
    return node is not None and node.ftag in {layer1.EdgeTags.Process, layer1.EdgeTags.State}


def move_node(node, new_parent, tag=None):
    for edge in node.incoming:
        if edge.parent == fparent(edge):
            copy_edge(edge, parent=new_parent, tag=tag)
            remove(edge.parent, edge)
            break
    # for (parent_id, child_id), count in Counter((edge.parent.ID, edge.child.ID) for edge in new_parent).items():
    #     if count > 1:
    #         raise ValueError("There are %d edges from %s to %s" % (count, parent_id, child_id))


AUX = {"have", "be", "will", "to", "do", "'s", "'ve", "'ll", "'re", "'d", "'m"}


def extract_aux(terminal, parent, grandparent):
    if get_annotation(terminal, Attr.LEMMA) in AUX and is_main_relation(grandparent) and (
            parent.ftag == layer1.EdgeTags.Function or
            parent.ftag in {layer1.EdgeTags.Elaborator, layer1.EdgeTags.Relator} and
            get_annotation(terminal, Attr.DEP) in {"aux", "auxpass"}):
        move_node(parent, fparent(grandparent), tag=layer1.EdgeTags.Function)
        return True
    return False


LIGHT_VERBS = {"take", "make", "give", "have", "pay"}


def set_light_verb_function(terminal, parent, grandparent):
    if get_annotation(terminal, Attr.LEMMA) in LIGHT_VERBS and \
            is_main_relation(grandparent) and parent.ftag == layer1.EdgeTags.Elaborator:
        if len(grandparent.centers) != 1 or len(grandparent.centers[0].get_terminals()) != 1 or \
                get_annotation(grandparent.centers[0].get_terminals()[0], Attr.DEP) != "ccomp":
            move_node(parent, grandparent, tag=layer1.EdgeTags.Function)
            return True
    return False


MODALS = {"can", "could", "may", "might", "shall", "should", "would", "must"}


def extract_modal(terminal, parent, grandparent):
    if (get_annotation(terminal, Attr.LEMMA) in MODALS or
        get_annotation(terminal, Attr.POS) in {"VERB", "ADV"} and
        get_annotation(terminal, Attr.DEP) not in {"aux", "auxpass"}) and \
            is_main_relation(grandparent) and parent.ftag == layer1.EdgeTags.Elaborator:
        move_node(parent, fparent(grandparent), tag=layer1.EdgeTags.Adverbial)
        return True
    return False


def extract_relator(terminal, parent, grandparent):
    following_uncle = None
    for node in grandparent.iter():
        if node.tag == layer1.NodeTags.Foundational and node.start_position == 1 + terminal.position and \
                node.ftag in {layer1.EdgeTags.Participant, layer1.EdgeTags.Adverbial}:
            following_uncle = node
    if following_uncle is not None and is_main_relation(grandparent) and parent.ftag == layer1.EdgeTags.Relator and \
            grandparent.end_position == terminal.position:
        move_node(parent, following_uncle)
        return True
    return False


def extract_that(terminal, parent, grandparent):
    del grandparent
    if get_annotation(terminal, Attr.LEMMA) == "that":
        following_scene = None
        for node in terminal.root.layer(layer1.LAYER_ID).heads[0].iter():
            if node.tag == layer1.NodeTags.Foundational and node.start_position == 1 + terminal.position and \
                    node.ftag == layer1.EdgeTags.ParallelScene:
                following_scene = node
        if following_scene is not None:
            move_node(parent, following_scene, tag=layer1.EdgeTags.Relator)
            return True
    return False


GROUND = {"seem", "feel", "sound", "taste", "look", "smell"}


def extract_ground(terminal, parent, grandparent):
    if get_annotation(terminal, Attr.LEMMA) in GROUND:
        if is_main_relation(grandparent) and parent.ftag == layer1.EdgeTags.Elaborator:
            move_node(parent, fparent(grandparent), tag=layer1.EdgeTags.Ground)
            return True
    return False


def fix_punct(terminal, parent, grandparent):
    del grandparent
    if parent.tag == layer1.NodeTags.Foundational and terminal.tag == layer0.NodeTags.Punct and not \
            (len(parent.terminals) + len(parent.punctuation) == len(parent.children) > 1):  # Unanalyzable
        for edge in parent:
            parent.remove(edge)
        parent.root.layer(layer1.LAYER_ID).add_punct(parent, terminal)
        return True
    return False


def fix_root_terminal_child(terminal, parent, grandparent):
    del grandparent
    if not parent.incoming:
        f1 = parent.root.layer(layer1.LAYER_ID).add_fnode(parent, layer1.EdgeTags.Function)
        move_node(terminal, f1)
        return True
    return False


def fix_unary_participant(terminal, parent, grandparent):
    del terminal, parent
    while grandparent.incoming:
        if len(grandparent.outgoing) == 1 and grandparent.outgoing[0].tag == layer1.EdgeTags.Participant:
            for edge in grandparent.incoming:
                copy_edge(edge, child=grandparent.children[0])
            destroy(grandparent)
            return True
        grandparent = grandparent.parents[0]
    return False


def flag_relator_starts_main_relation(terminal, parent, grandparent):
    return grandparent.start_position == terminal.position and \
        is_main_relation(grandparent) and parent.ftag == layer1.EdgeTags.Relator


def flag_suspected_secondary(terminal, parent, grandparent):
    return get_annotation(terminal, Attr.DEP) not in {"det"} and \
        is_main_relation(grandparent) and parent.ftag == layer1.EdgeTags.Elaborator


RULES = (extract_ground, extract_aux, set_light_verb_function, extract_modal, extract_relator, extract_that,
         fix_punct, fix_root_terminal_child, fix_unary_participant,
         flag_relator_starts_main_relation, flag_suspected_secondary)


def convert_passage(passage, report_writer):
    for rule in RULES:
        for terminal in passage.layer(layer0.LAYER_ID).all:
            parent = fparent(terminal)
            grandparent = fparent(parent)
            grandparent_str = str(grandparent)
            if len(parent.children) == 1 and rule(terminal, parent, grandparent):
                report_writer.writerow((rule.__name__, passage.ID, terminal.ID, get_annotation(terminal, Attr.POS),
                                        grandparent_str, fparent(fparent(terminal))))


def main(args):
    textutil.BATCH_SIZE = 1
    os.makedirs(args.outdir, exist_ok=True)
    with open(args.outfile, "w", encoding="utf-8", newline="") as f:
        writer = csv.writer(f)
        writer.writerow(("rule", "passage", "terminal", "pos", "before", "after"))
        for passage in annotate_all(get_passages_with_progress_bar(args.passages, desc="Converting"),
                                    verbose=args.verbose):
            convert_passage(passage, report_writer=writer)
            write_passage(passage, outdir=args.outdir, prefix=args.prefix, verbose=args.verbose)
            f.flush()
    print("Wrote '%s'" % args.outfile)


if __name__ == "__main__":
    argparser = ArgumentParser(description=desc)
    argparser.add_argument("passages", nargs="+", help="the corpus, given as xml/pickle file names")
    argparser.add_argument("-o", "--outdir", default=".", help="output directory")
    argparser.add_argument("-p", "--prefix", default="", help="output filename prefix")
    argparser.add_argument("-O", "--outfile", default=os.path.splitext(argparser.prog)[0] + ".csv", help="log file")
    argparser.add_argument("-v", "--verbose", action="store_true", help="print tagged text for each passage")
    main(argparser.parse_args())



import argparse
import os
import sqlite3
from glob import glob
from xml.etree.ElementTree import ElementTree, fromstring

import ucca.convert
from ucca.ioutil import write_passage

desc = """Parses an XML in UCCA site format.

The input can be given as either an XML file or a DB file with passage ID
and user name, and the output is either the standard format XML or
a pickled object.
Possible input methods are using a DB file with pid and user, which gets the
annotation of the specified user for the specified passage from teh DB file,
or using filenames of a site-formatted XML file.

"""


def site2passage(filename):
    """Opens a file and returns its parsed Passage object"""
    with open(filename, encoding="utf-8") as f:
        print("Reading '%s'..." % filename)
        return ucca.convert.from_site(ElementTree().parse(f))


def db2passage(handle, pid, user):
    """Gets the annotation of user to pid from the DB handle - returns a passage"""
    handle.execute("SELECT id FROM users WHERE username=?", (user,))
    uid = handle.fetchone()[0]
    handle.execute("SELECT xml FROM xmls WHERE paid=? AND uid=? ORDER BY ts DESC", (pid, uid))
    return ucca.convert.from_site(fromstring(handle.fetchone()[0]))


def main(args):
    os.makedirs(args.out_dir, exist_ok=True)
    for filename, passage in ((filename, site2passage(filename)) for pattern in args.filenames
                              for filename in sorted(glob(pattern)) or [pattern]) if args.filenames \
            else ((pid, db2passage(sqlite3.connect(args.db).cursor(), pid, args.user)) for pid in args.pids):
        write_passage(passage, outdir=args.out_dir, binary=args.binary)


def check_illegal_combinations(args):
    if args.db and not (args.pids and args.user):
        argparser.error("Must specify a username and a passage ID when using DB file option")
    if (args.pids or args.user) and not args.db:
        argparser.error("Cannot use user and passage ID options without DB file")
    return args


if __name__ == "__main__":
    argparser = argparse.ArgumentParser(description=desc)
    argparser.add_argument("filenames", nargs="*", help="XML file name to convert")
    argparser.add_argument("-d", "--db", help="DB file to get input from")
    argparser.add_argument("-o", "--out-dir", default=".", help="output directory for standard XML")
    argparser.add_argument("-b", "--binary", help="output file for binary pickle")
    argparser.add_argument("-p", "--pids", nargs="*", type=int, help="PassageIDs to query DB")
    argparser.add_argument("-u", "--user", help="Username to DB query")
    main(check_illegal_combinations(argparser.parse_args()))



import argparse
import re
import sys
from glob import glob
from itertools import groupby
from operator import attrgetter

from tqdm import tqdm

from ucca import layer0
from ucca.ioutil import get_passages_with_progress_bar


def gen_lines(filenames):
    for filename in glob(filenames) or [filenames]:
        with open(filename, encoding="utf-8") as f:
            try:
                for line in map(str.strip, f):
                    if line and not line.startswith("#"):
                        yield re.sub(r"\[\d+\]", "", line)  # Remove numbers inside brackets
            except UnicodeDecodeError as e:
                raise IOError("Failed reading '%s'" % filename) from e


class CandidateMatcher:
    def __init__(self, text):
        self.text = text
        self.char_map = {}
        no_space_chars = []
        for i, char in enumerate(text):
            if not char.isspace():
                self.char_map[len(no_space_chars)] = i
                no_space_chars.append(char)
        self.no_space_text = "".join(no_space_chars)

    def __call__(self, no_space_text):
        try:
            index = self.no_space_text.index(no_space_text)
            return self.text[self.char_map[index]:self.char_map[index + len(no_space_text) - 1] + 1]
        except ValueError:
            return None


def match_passage_text(passage, matchers, out):
    passage_tokens = sorted(passage.layer(layer0.LAYER_ID).all, key=attrgetter("position"))
    for paragraph, terminals in groupby(passage_tokens, key=attrgetter("paragraph")):
        tokens = [terminal.text for terminal in terminals]
        no_space_text = "".join(tokens)
        match = next(filter(None, (matcher(no_space_text) for matcher in matchers)), "@@@" + " ".join(tokens))
        print(passage.ID, match, sep="\t", file=out)


def alternative_spellings(text):
    yield text


def main(args):
    matchers = [CandidateMatcher(spelling) for line in tqdm(list(gen_lines(args.text)),
                                                            desc="Indexing " + args.text, unit=" lines")
                for spelling in alternative_spellings(line)]
    out = open(args.out, "w", encoding="utf-8") if args.out else sys.stdout
    for p in get_passages_with_progress_bar(args.filenames, desc="Matching", converters={}):
        match_passage_text(p, matchers, out)
    out.close()


if __name__ == "__main__":
    argparser = argparse.ArgumentParser(description="Match UCCA passages to original text and print aligned lines")
    argparser.add_argument("text", help="file of text to match to")
    argparser.add_argument("filenames", nargs="+", help="files or directories of UCCA passages to match")
    argparser.add_argument("-o", "--out", default="text.tsv", help="output file")
    argparser.add_argument("-l", "--lang", default="en", help="spaCy language")
    main(argparser.parse_args())



import os
from argparse import ArgumentParser

from ucca import visualization, layer0
from ucca.convert import split2sentences
from ucca.ioutil import get_passages, get_passages_with_progress_bar, external_write_mode


def print_text(args, text, suffix):
    if args.out_dir:
        with open(os.path.join(args.out_dir, suffix), "w") as f:
            print(text, file=f)
    else:
        with external_write_mode():
            print(text)


def main(args):
    if args.out_dir:
        os.makedirs(args.out_dir, exist_ok=True)
        if not args.tikz:
            import matplotlib
            matplotlib.use('Agg')
    to_stdout = (args.tikz or args.standoff) and not args.out_dir
    t = args.passages
    t = get_passages(t) if to_stdout else get_passages_with_progress_bar(t, desc="Visualizing")
    if args.sentences:
        t = (sentence for passage in t for sentence in split2sentences(passage))
    for passage in t:
        if args.tikz:
            print_text(args, visualization.tikz(passage), passage.ID + ".tikz.txt")
        elif args.standoff:
            print_text(args, visualization.standoff(passage), passage.ID + ".ann")
        else:
            import matplotlib.pyplot as plt
            width = len(passage.layer(layer0.LAYER_ID).all) * 19 / 27
            plt.figure(passage.ID, figsize=(width, width * 10 / 19))
            visualization.draw(passage, node_ids=args.node_ids)
            if args.out_dir:
                plt.savefig(os.path.join(args.out_dir, passage.ID + "." + args.format))
                plt.close()
            else:
                plt.show()


if __name__ == "__main__":
    argparser = ArgumentParser(description="Visualize the given passages as graphs.")
    argparser.add_argument("passages", nargs="+", help="UCCA passages, given as xml/pickle file names")
    group = argparser.add_mutually_exclusive_group()
    group.add_argument("-t", "--tikz", action="store_true", help="print tikz code rather than showing plots")
    group.add_argument("-s", "--standoff", action="store_true", help="print standoff code rather than showing plots")
    argparser.add_argument("-o", "--out-dir", help="directory to save figures in (otherwise displayed immediately)")
    argparser.add_argument("-i", "--node-ids", action="store_true", help="print tikz code rather than showing plots")
    argparser.add_argument("-f", "--format", choices=("png", "svg"), default="png", help="image format")
    argparser.add_argument("--sentences", help="split to sentences to avoid huge plots")
    main(argparser.parse_args())



#!/usr/bin/env python3
import argparse
import json
import sys

from tqdm import tqdm

from ucca import normalization, validation
from ucca.convert import from_json
from ucca.ioutil import write_passage
from uccaapp.api import ServerAccessor

desc = """Download task from UCCA-App and convert to a passage in standard format"""


class TaskDownloader(ServerAccessor):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def download_tasks(self, task_ids, by_filename=False, validate=None, log=None, **kwargs):
        if by_filename:
            task_ids_from_file = []
            for filename in task_ids:
                with open(filename, 'r') as f:
                    task_ids_from_file += list(filter(None, map(str.strip, f)))
            task_ids = task_ids_from_file
        validate_h = open(validate, "w", encoding="utf-8") if validate else None
        log_h = open(log, "w", encoding="utf-8") if log else None
        for task_id in tqdm(task_ids, unit=" tasks", desc="Downloading"):
            yield self.download_task(task_id, validate=validate_h, log=log_h, **kwargs)
        if validate:
            validate_h.close()
        if log:
            log_h.close()

    def download_task(self, task_id, normalize=False, write=True, validate=None, binary=None, log=None, out_dir=None,
                      prefix=None, by_external_id=False, verbose=False, write_valid_only=False, strict=False, **kwargs):
        del kwargs
        task = self.get_user_task(task_id)
        user_id = task["user"]["id"]
        passage = None
        try:
            passage = next(iter(from_json(task, by_external_id=by_external_id)))
        except ValueError as e:
            if strict:
                raise ValueError("Failed reading json for task %s:\n%s" % (task_id, json.dumps(task))) from e
            print("", task_id, user_id, "Failed reading json", file=validate or sys.stderr, sep="\t", flush=True)
        if normalize and passage is not None:
            try:
                normalization.normalize(passage)
            except AssertionError as e:
                if strict:
                    raise ValueError("Failed normalizing task %s:\n%s" % (task_id, json.dumps(task))) from e
                print(passage.ID, task_id, user_id, "Failed normalizing task: %s" % e, file=validate or sys.stderr,
                      sep="\t", flush=True)
        if log:
            print(passage.ID, task_id, user_id, task["user_comment"], task["created_at"], task["updated_at"],
                  file=log, sep="\t", flush=True)
        ret = passage, task_id, user_id
        if validate or write_valid_only:
            for error in validation.validate(passage, linkage=False):
                if validate:
                    print(passage.ID, task_id, user_id, error, file=validate, sep="\t", flush=True)
                if write_valid_only:
                    return ret
        if write:
            write_passage(passage, binary=binary, outdir=out_dir, prefix=prefix, verbose=verbose)
        return ret

    @staticmethod
    def add_arguments(argparser):
        argparser.add_argument("task_ids", nargs="+", help="IDs of tasks to download and convert")
        argparser.add_argument("-f", "--by-filename", action="store_true", help="treat task_ids as a filename, "
                                                                                "otherwise it is a list of IDs")
        TaskDownloader.add_write_arguments(argparser)
        argparser.add_argument("-V", "--validate", help="run validation on downloaded passages and save errors to file")
        argparser.add_argument("-N", "--normalize", action="store_true", help="normalize downloaded passages")
        argparser.add_argument("--strict", action="store_true", help="fail on reading or normalization error")
        argparser.add_argument("-l", "--log", help="filename to write log of downloaded passages to")
        ServerAccessor.add_arguments(argparser)

    @staticmethod
    def add_write_arguments(argparser):
        argparser.add_argument("-o", "--out-dir", default=".", help="output directory")
        argparser.add_argument("-p", "--prefix", default="", help="output filename prefix")
        argparser.add_argument("-x", "--by-external-id", action="store_true", help="save filename by external ID")
        argparser.add_argument("-b", "--binary", action="store_true", help="write in binary format (.pickle)")
        argparser.add_argument("-n", "--no-write", action="store_false", dest="write", help="do not write files")
        argparser.add_argument("--write-valid-only", action="store_true", help="only write passages that passed "
                                                                               "validation")


def main(**kwargs):
    list(TaskDownloader(**kwargs).download_tasks(**kwargs))


if __name__ == "__main__":
    argument_parser = argparse.ArgumentParser(description=desc)
    TaskDownloader.add_arguments(argument_parser)
    main(**vars(argument_parser.parse_args()))
    sys.exit(0)



from ucca import layer0, layer1
from ucca.layer0 import NodeTags as L0Tags
from ucca.layer1 import EdgeTags as ETags, NodeTags as L1Tags

NO_MULTIPLE_INCOMING_CATEGORIES = {ETags.Function, ETags.ParallelScene, ETags.Linker, ETags.LinkRelation,
                                   ETags.Connector, ETags.Punctuation, ETags.Terminal}
TOP_CATEGORIES = {ETags.ParallelScene, ETags.Linker, ETags.Function, ETags.Ground, ETags.Punctuation,
                  ETags.LinkRelation, ETags.LinkArgument, ETags.Connector}
COORDINATED_MAIN_REL = "Coordinated_Main_Rel."


def traverse_up_centers(node):
    while True:
        found_center = False
        for edge in node.incoming:
            if not edge.attrib.get("remote") and layer1.EdgeTags.Center in edge.tags:
                node = edge.parent
                found_center = True
                continue
        if not found_center:
            return node


def fparent(node_or_edge):
    try:
        return node_or_edge.fparent
    except AttributeError:
        try:
            return node_or_edge.parent
        except AttributeError:
            return node_or_edge.parents[0] if node_or_edge.parents else None


def remove_unmarked_implicits(node):
    while node is not None and not node.children and not node.attrib.get("implicit"):
        parent = fparent(node)
        if parent is None:
            break
        node.destroy()
        node = parent


def remove(parent, child):
    if parent is not None:
        parent.remove(child)
        remove_unmarked_implicits(parent)


def destroy(node_or_edge):
    parent = fparent(node_or_edge)
    try:
        node_or_edge.destroy()
    except AttributeError:
        parent.remove(node_or_edge)
    if parent is not None:
        remove_unmarked_implicits(parent)
    return parent


def copy_edge(edge, parent=None, child=None, tag=None, attrib=None):
    if parent is None:
        parent = edge.parent
    if child is None:
        child = edge.child
    if not tag:
        categories = [(c.tag, c.slot, c.layer, c.parent) for c in edge.categories]
    else:
        categories = [(tag,)]
    if attrib is None:
        attrib = edge.attrib
    if parent in child.iter():
        # raise ValueError("Created cycle (%s->%s) when trying to normalize '%s'" % (
        #     "->".join(n.ID for n in child.iter() if parent in n.iter()), child.ID, parent))
        return False
    parent.add_multiple(categories, child, edge_attrib=attrib)
    return True


def replace_center(edge):
    if len(edge.parent) == 1 and not edge.parent.parents:
        return ETags.ParallelScene
    if edge.parent.participants and not edge.parent.is_scene():
        return ETags.Process  # TODO should be state if the word is a copula
    return edge.tag


def replace_edge_tags(node):
    for edge in node:
        if not edge.attrib.get("remote") and edge.tag == ETags.Center:
            edge.tag = replace_center(edge)
        elif node.parallel_scenes:
            if edge.tag == ETags.Connector:
                edge.tag = ETags.Linker
        elif edge.tag == ETags.Linker:
            edge.tag = ETags.Connector
        elif node.is_scene():
            if edge.tag == ETags.Elaborator:
                edge.tag = ETags.Adverbial
        elif edge.tag == ETags.Adverbial:
            edge.tag = ETags.Elaborator


def move_elements(node, tags, parent_tags, forward=True):
    for edge in node:
        if edge.child.tag == L1Tags.Foundational and edge.tag in ((tags,) if isinstance(tags, str) else tags):
            try:
                parent_edge = min((e for e in node if e != edge and e.child.tag == L1Tags.Foundational),
                                  key=lambda e: abs(((edge.child.start_position - e.child.end_position),
                                                     (e.child.start_position - edge.child.end_position))[forward]))
            except ValueError:
                continue
            if parent_edge.tag in ((parent_tags,) if isinstance(parent_tags, str) else parent_tags):
                parent = parent_edge.child
                if copy_edge(edge, parent=parent):
                    remove(node, edge)


def move_scene_elements(node):
    if node.parallel_scenes:
        move_elements(node, tags=(ETags.Relator, ETags.Elaborator, ETags.Center), parent_tags=ETags.ParallelScene)


def move_sub_scene_elements(node):
    if node.is_scene():
        move_elements(node, tags=(ETags.Elaborator, ETags.Center), parent_tags=ETags.Participant, forward=False)


def separate_scenes(node, l1, top_level=False):
    if (node.is_scene() or node.participants) and (top_level or node.parallel_scenes):
        edges = list(node)
        scene = l1.add_fnode(node, ETags.ParallelScene)
        for edge in edges:
            if edge.tag not in (ETags.ParallelScene, ETags.Punctuation, ETags.Linker, ETags.Ground):
                if copy_edge(edge, parent=scene):
                    remove(node, edge)


def lowest_common_ancestor(*nodes):
    parents = [nodes[0]] if nodes else []
    while parents:
        for parent in parents:
            if parent.tag == L1Tags.Foundational and (not parent.terminals or nodes[1:]) \
                    and all(n in parent.iter() for n in nodes[1:]):
                return parent
        parents = [p for n in parents for p in n.parents]
    return None


def nearest_word(l0, position, step):
    while True:
        position += step
        try:
            terminal = l0.by_position(position)
        except IndexError:
            return None
        if terminal.tag == L0Tags.Word:
            return terminal


def nearest_parent(l0, *terminals):
    return lowest_common_ancestor(*filter(None, (nearest_word(l0, terminals[0].position, -1),
                                                 nearest_word(l0, terminals[-1].position, 1))))


def reattach_punct(l0, l1):
    detach_punct(l1)
    attach_punct(l0, l1)


def attach_punct(l0, l1):
    for terminal in l0.all:
        if layer0.is_punct(terminal) and not terminal.incoming:
            l1.add_punct(nearest_parent(l0, terminal), terminal)


def detach_punct(l1):
    for node in l1.all:
        if node.tag == L1Tags.Punctuation:
            destroy(node)


def reattach_terminals(l0, l1):
    attach_terminals(l0, l1)
    for terminal in l0.all:
        for edge in terminal.incoming:
            if any(e.tag != ETags.Terminal for e in edge.parent):
                node = l1.add_fnode(edge.parent, ETags.Center)
                if copy_edge(edge, parent=node):
                    remove(edge.parent, edge)


def attach_terminals(l0, l1):
    for terminal in l0.all:
        if not terminal.incoming:
            node = l1.add_fnode(nearest_parent(l0, terminal), ETags.Function)
            node.add(ETags.Terminal, terminal)


def flatten_centers(node):
    """
    Whenever there are Cs inside Cs, remove the external C.
    Whenever there is a C as an only child, remove it.
    """
    if node.tag == L1Tags.Foundational and len(node.centers) == 1:
        if node.ftag == ETags.Center and len(fparent(node).centers) == 1:  # Center inside center
            for edge in node.incoming:
                if edge.attrib.get("remote"):
                    copy_edge(edge, child=node.centers[0])
            for edge in node:
                copy_edge(edge, parent=fparent(node))
            return destroy(node)
        elif len(node.children) == 1:  # Center as only child
            for edge in node.incoming:
                attrib = edge.attrib
                if node.outgoing[0].attrib.get("remote"):
                    attrib["remote"] = True
                copy_edge(edge, child=node.centers[0], attrib=attrib)
            return destroy(node)
    return node


def flatten_functions(node):
    """
    Whenever there is an F as an only child, remove it. If an F has non-terminal children, move them up.
    """
    if node.tag == L1Tags.Foundational and node.incoming:  # Avoid creating root->terminal edge
        for child in node.functions:
            if len(child.children) > len(child.terminals):
                for edge in child:
                    copy_edge(edge, parent=node, tag=ETags.Function if edge.tag == ETags.Center else edge.tag)
                destroy(child)
        if len(node.functions) == len(node.children) == 1:
            for edge in node.incoming:
                copy_edge(edge, child=node.functions[0])
            return destroy(node)
    return node


def flatten_participants(node):
    """
    Whenever there is an A as an only child, remove it.
    If there is an implicit A in a scene without a main relation, remove it.
    """
    if node.tag == L1Tags.Foundational:
        participants = node.participants
        if len(participants) == len(node.children) == 1 and len(participants[0].ftags) == 1:
            for edge in node.incoming:
                copy_edge(edge, child=participants[0])
            return destroy(node)
        elif participants and not node.is_scene():
            for child in participants:
                if child.attrib.get("implicit"):
                    destroy(child)
    return node


def split_coordinated_main_rel(node, l1):
    for edge in node:
        attrib = edge.attrib.copy()
        if attrib.pop(COORDINATED_MAIN_REL, None):
            assert {ETags.Process, ETags.State}.intersection(edge.tags), \
                "%s node without main relation: %s" % (COORDINATED_MAIN_REL, node)
            main_rel = edge.child
            centers = main_rel.centers
            assert centers, "%s node without centers: %s" % (COORDINATED_MAIN_REL, main_rel)
            main_rel_non_centers = [e for e in main_rel.outgoing if ETags.Center not in e.tags]
            main_rel_incoming = list(main_rel.incoming)
            main_rel.destroy()
            top = fparent(node)
            if ETags.ParallelScene in node.ftags:
                top.remove(node)
            else:
                top = node
            outgoing = list(node.outgoing)
            scenes = []
            for center in centers:
                new_scene = l1.add_fnode(top, ETags.ParallelScene)
                copy_edge(edge, parent=new_scene, child=center, attrib=attrib)
                for scene_edge in outgoing:
                    if scene_edge.ID != edge.ID and not (
                            scenes and NO_MULTIPLE_INCOMING_CATEGORIES.intersection(scene_edge.tags)):
                        # Not the CMR edge itself, and not a category that does not allow multiple parents
                        copy_edge(scene_edge, parent=new_scene, attrib={"remote": True} if scenes else None)
                scenes.append(new_scene)
            for main_rel_edge in main_rel_non_centers:
                tags = main_rel_edge.tags
                copy_edge(main_rel_edge, parent=top if TOP_CATEGORIES.issuperset(tags) else scenes[0],
                          tag=ETags.Linker if ETags.Connector in main_rel_edge.tags else None)
            for scene_edge in outgoing:
                if scene_edge.ID != edge.ID:
                    destroy(scene_edge)
            for remote_edge in main_rel_incoming:
                if remote_edge.attrib.get("remote"):
                    copy_edge(remote_edge, child=centers[0])
            if not node.incoming:
                node.destroy()
    return node


def normalize_node(node, l1, extra):
    if node.tag == L1Tags.Foundational:
        if extra:
            replace_edge_tags(node)
            move_scene_elements(node)
            move_sub_scene_elements(node)
        separate_scenes(node, l1, top_level=node in l1.heads)
        node = split_coordinated_main_rel(node, l1)
        if node is None:
            return None
        node = flatten_centers(node)
        if node is None:
            return
        node = flatten_functions(node)
        if node is None:
            return
        flatten_participants(node)


def normalize(passage, extra=False):
    l0 = passage.layer(layer0.LAYER_ID)
    l1 = passage.layer(layer1.LAYER_ID)
    reattach_punct(l0, l1)
    heads = list(l1.heads)
    stack = [heads]
    visited = set()
    path = []
    path_set = set()
    while stack:
        for edge in stack[-1]:
            try:
                node = edge.child
            except AttributeError:
                node = edge
            if node in path_set:
                destroy(edge)
            elif node not in visited:
                visited.add(node)
                path.append(node)
                path_set.add(node)
                stack.append(node)
                normalize_node(node, l1, extra)
                break
        else:
            if path:
                path_set.remove(path.pop())
            stack.pop()
    reattach_punct(l0, l1)
    if extra:
        reattach_terminals(l0, l1)
